﻿local metadata = {
"## Interface: FS17 1.5.3.1",
"## Title: UnitConvert",
"## Notes: Allows Farming Simulator 17 to display custom units of measurement",
"## Author: ThundR75",
"## Version: 1.2.2.1",
"## Date: 09.05.2018"
}
--[[
## ChangeLog
	Version 1.2.2.1
		- Added support for FarmingTablet App: AnimalStatus
		- Added support for FarmingTablet App: Horsch Management
		- Added support for FarmingTablet App: FarmSiloSystem
		- Added support for FarmingTablet App: StorageHouse
		- Changed "Storage Capacity" to "Silo Capacity" on statistics screen.
		- Moved DriveControl nearest neighbor display.
		
	Version 1.2.2.0
		- New third party mod support system added.
		- UnitConvert now works with the Farming Tablet mod.

	Version 1.2.1.1
		- Mass factor is now working correctly.

	Version 1.2.1.0
		- Fuel usage display now works correctly in multiplayer.
		- Yield display now works correctly in multiplayer.
		- Yield display can be toggled between "vehicle", "price" and "off".
		- Adding a unitConvert.xml file to the base folder of your map allows adding/changing any value.
		- Editing the configs/unitConvert_<lang>.xml file allows you to add/change any value.
		- Fill type groups are now displayed in the UnitConvert menu.
		- UnitConvert now works with the GPS mod.

	Version 1.2.0.3
		- Added internal fill level hud to list of converted displays.

	Version 1.2.0.2
		- Fixed bug where new fill types weren't saving their configuration.

	Version 1.2.0.1
		- Added help (readme.txt) to the in-game menu.

	Version 1.2.0.0
		- Added ability for modders to add UnitConvert data to their mods/maps.
		- Unit precision is now adjustable via the UnitConvert menu.
		- Mass factor is now adjustable via the UnitConvert menu.
		- Fill type categories have been replaced by fill type groups (internal).
		- Fill type group priority adjustable via the UnitConvert menu.
		- UnitConvert menu now scrolls, supporting large lists.
		- Weights can now be displayed anywhere.
		- Fill type names are now displayed on prices screen.
		- Fill type names are now displayed on the vehicle hud.
		- Code cleanup
		- Bug Fixes
]]--

local debugLvl = 0
local uiScale = g_gameSettings:getValue("uiScale")
local isDev = false

-- [LOCAL FUNCTIONS] **********************************************************************************************************************************************************************************

local function getmdata(v)
	v = "## "..v.. ": "
	for i = 1, #metadata do
		local _, n = string.find(metadata[i], v)
		if n then
			return (string.sub (metadata[i], n + 1))
		end
	end
end

local function getText(text, ...)
	return string.format(g_i18n:getText(text), ...)
end

local function raiseError(text, showCallstack, ...)
	if (text and type(text) ~= "string") or (showCallstack and type(showCallStack) ~= "boolean") then
		raiseError()
		return
	end

	if not text and showCallstack == nil then
		showCallstack = true
	end

	text = Utils.getNoNil(text, "UC_error_internal")
	text = tostring(text)

	if g_i18n:hasText(text) then
		text = getText(text, ...)
	end

	if showCallstack then
		printCallstack()
	end

	error("**** "..string.format("%s %s: %s", getmdata("Title"), getmdata("Version"), text), 0)
end

local function xAssert(arg, text, showCallstack)
	if not arg then
		raiseError(text, showCallstack)
		return
	end
	return arg
end

local function debugMsg(lvl, text, ...)
	if tonumber(lvl) == nil or type(text) ~= "string" then
		raiseError()
		return
	end

	text = tostring(text)
	if g_i18n:hasText(text) then
		text = getText(text, ...)
	end

	if lvl <= debugLvl then
		print("**** "..getmdata("Title").." v"..getmdata("Version")..": "..text)
	end
end

function getIsBoolean(arg)
	if type(arg) == "boolean" then
		return true
	end
end

function getIsString(arg)
	if type(arg) == "string" then
		return true
	end
end

function getIsNumber(arg)
	if tonumber(arg) then
		return true
	end
end

function getIsTable(arg)
	if type(arg) == "table" then
		return true
	end
end

local function sortTable(tbl, key)
	if not getIsTable(tbl) or key == nil or (not getIsNumber(key) and not getIsString(key)) then
		raiseError()
		return
	end

	key = Utils.getNoNil(tonumber(key), key)
	local maxVal = 0

	local tblTemp = {}
	for k, v in pairs(tbl) do
		local idx = tonumber(v[key])
		if idx then
			idx = idx * #tbl
			while tblTemp[idx] do
				idx = idx + 1
			end
			tblTemp[idx] = v
			maxVal = math.max(maxVal, idx)
		end
	end

	local tblSorted = {}
	for idx = 1, maxVal do
		if tblTemp[idx] then
			tblTemp[idx][key] = #tblSorted + 1
			table.insert(tblSorted, tblTemp[idx])
		end
	end

	return tblSorted
end

-- [DECLARATIONS] *************************************************************************************************************************************************************************************

UnitConvert = {}
UnitConvert.version = getmdata("Version")
UnitConvert.modDir = g_currentModDirectory
UnitConvert.addOns = {}

getfenv(0)["g_unitConvert"] = UnitConvert
local ucMod = g_unitConvert

-- Third Party AddOn Scripts
ucMod.addOnDir = ucMod.modDir.."addons/"
source(ucMod.addOnDir.."ucGPSMod.lua")
source(ucMod.addOnDir.."ucFarmingTablet.lua")
source(ucMod.addOnDir.."ucFarmingTabletAnimalStatus.lua")
source(ucMod.addOnDir.."ucFarmingTabletBunkerSiloOverview.lua")
source(ucMod.addOnDir.."ucFarmingTabletStorageOverview.lua")
source(ucMod.addOnDir.."ucFarmingTabletFactoryExtension.lua")
source(ucMod.addOnDir.."ucFarmingTabletFarmSiloSystem.lua")
source(ucMod.addOnDir.."ucFarmingTabletStorageHouse.lua")
source(ucMod.addOnDir.."ucFarmingTabletHorsch.lua")
source(ucMod.addOnDir.."ucDriveControl.lua")

-- [GLOBAL FUNCTIONS] *********************************************************************************************************************************************************************************

function UnitConvert:strToNum(str)
	if str == nil or (not getIsNumber(str) and not getIsString(str)) then
		raiseError()
		return
	end

	str = tostring(str)
	str = string.gsub(str, ",", "")
	str = string.match(str, "%-?%d+%.?%d*[eE]?%-?%d*")

	return tonumber(str)
end

function UnitConvert:numToStr(num, arg, roundType, argIsPrecision, showCommas)
	if arg == nil then arg = 0 end

	if not getIsNumber(num) or not getIsNumber(arg)
	or (argIsPrecision and not getIsBoolean(argIsPrecision))
	or (showCommas and not getIsBoolean(showCommas))
	then
		raiseError()
		return
	end

	local precision = arg
	if not argIsPrecision then
		precision = arg - string.len(math.floor(num))
		precision = math.max(0, precision)
	end

	if roundType == "round" then
		num = math.floor((num * (10 ^ precision)) + 0.5) / (10 ^ precision)
	elseif roundType == "floor" then
		num = math.floor(num * (10 ^ precision)) / (10 ^ precision)
	else
		num = math.ceil(num * (10 ^ precision)) / (10 ^ precision)
	end

	local num = string.format("%."..tostring(precision).."f", num)

	if showCommas then
		local v = 0
		while true do
			num, v = string.gsub(num, "^(%-?%d+)(%d%d%d)", "%1,%2")
			if v == 0 then break end
		end
	end

	return tostring(num)
end

function UnitConvert:splitString(pattern, text)
	if not getIsString(pattern) or text == nil or (not getIsNumber(text) and not getIsString(text)) then
		raiseError()
		return
	end

	text = tostring(text)
	local retTbl = Utils.splitString(pattern, text)

	for idx = 1, #retTbl do
		retTbl[idx] = Utils.getNoNil(tonumber(retTbl[idx]), retTbl[idx])
	end

	return retTbl
end

function UnitConvert:getIsFruitTypeValid(fruitType)
	if fruitType then
		if fruitType ~= FruitUtil.FRUITTYPE_UNKNOWN then
			if xAssert(Utils.getNoNil(FruitUtil.fruitIndexToDesc[fruitType], FruitUtil.fruitTypes[fruitType])) then
				return true
			end
		end
	end
end

function UnitConvert:getIsFillTypeValid(fillType)
	if fillType then
		if fillType ~= FillUtil.FILLTYPE_UNKNOWN then
			if xAssert(Utils.getNoNil(FillUtil.fillTypeIndexToDesc[fillType], FillUtil.fillTypeNameToDesc[fillType])) then
				return true
			end
		end
	end
end

function UnitConvert:getIsFillTypeCategoryValid(fillTypeCategory)
	if fillTypeCategory then
		fillTypeCategory = Utils.getNoNil(FillUtil.fillTypeCategoryNameToInt[fillTypeCategory], fillTypeCategory)
		if fillTypeCategory ~= FillUtil.FILLTYPE_CATEGORY_UNKNOWN then
			if xAssert(FillUtil.fillTypeCategoryIntToName[fillTypeCategory]) then
				return true
			end
		end
	end
end

function UnitConvert:getUnit(unit)
	unit = xAssert(Utils.getNoNil(tonumber(unit), unit))
	return xAssert(Utils.getNoNil(ucMod.unitsByIndex[unit], ucMod.unitsByName[unit]))
end

function UnitConvert:getUnitType(unitType)
	unitType = xAssert(Utils.getNoNil(tonumber(unitType), unitType))
	return xAssert(Utils.getNoNil(ucMod.unitTypesByIndex[unitType], ucMod.unitTypesByName[unitType]))
end

function UnitConvert:getUnitSet(unitSet)
	unitSet = xAssert(Utils.getNoNil(tonumber(unitSet), unitSet))
	return xAssert(Utils.getNoNil(ucMod.unitSetsByIndex[unitSet], ucMod.unitSetsByName[unitSet]))
end

function UnitConvert:getFillTypeGroup(fillTypeGroup)
	fillTypeGroup = xAssert(Utils.getNoNil(tonumber(fillTypeGroup), fillTypeGroup))
	return xAssert(Utils.getNoNil(ucMod.fillTypeGroupsByIndex[fillTypeGroup], ucMod.fillTypeGroupsByName[fillTypeGroup]))
end

function UnitConvert:getFillType(fillType)
	fillType = xAssert(Utils.getNoNil(tonumber(fillType), fillType))
	return xAssert(Utils.getNoNil(ucMod.fillTypesByIndex[fillType], ucMod.fillTypesByName[fillType]))
end

function UnitConvert:getFillTypesByCategories(categories)
-- categories:	Giants categories to parse (table, string, or number - by name or index)
-- return:		All fill types belonging to the above categories

	if not categories then
		raiseError()
		return
	end

	local fillTypes = {}
	if not getIsTable(categories) then categories = ucMod:splitString(" ", categories) end

	for _, category in pairs(categories) do
		if ucMod:getIsFillTypeCategoryValid(category) then
			category = xAssert(Utils.getNoNil(tonumber(category), FillUtil.fillTypeCategoryNameToInt[category]))
			local fillTypeList = FillUtil.categoryToFillTypes[category]
			if fillTypeList then
				for fillType, isValid in pairs(fillTypeList) do
					if isValid and ucMod:getIsFillTypeValid(fillType) and not fillTypes[fillType] then
						table.insert(fillTypes, fillType)
					end
				end
			end
		end
	end

	if #fillTypes > 0 then return fillTypes end
end

function UnitConvert:getHeaviestFillType(fillTypes)
-- fillTypes: 	Fill types to check (table, number, or string - by name or index)
-- return:		Heaviest fill type among the above list

	if not fillTypes then
		raiseError()
		return
	end

	if not getIsTable(fillTypes) then fillTypes = ucMod:splitString(" ", fillTypes) end

	local maxMass = 0
	local retFillType = FillUtil.FILLTYPE_UNKNOWN

	for _, fillType in pairs(fillTypes) do
		if ucMod:getIsFillTypeValid(fillType) then
			fillType = ucMod:getFillType(fillType)
			local fillTypeMass = fillType.massPerLiter * fillType.massFactor
			if fillTypeMass >= maxMass then
				maxMass = fillTypeMass
				retFillType = fillType.index
			end
		end
	end

	if retFillType and retFillType ~= FillUtil.FILLTYPE_UNKNOWN then
		return retFillType
	end
end

function UnitConvert:getUnitTypesByGroup(fillTypeGroup)
	fillTypeGroup = ucMod:getFillTypeGroup(fillTypeGroup)
	return xAssert(fillTypeGroup.unitTypes)
end

function UnitConvert:getAreaUnit()
	return ucMod:getUnit(ucMod.areaUnit)
end

function UnitConvert:getLengthUnit()
	return ucMod:getUnit(ucMod.lengthUnit)
end

function UnitConvert:getWeightUnit()
	return ucMod:getUnit(ucMod.weightUnit)
end

function UnitConvert:getFillUnit(fillTypes, unitSet)
-- fillTypes:	list of fillTypes to convert (single, list, or table - by index or name)
-- unitSet:		unit set to retrieve (by index or name)
-- return:		unit (table)

	if (not getIsTable(fillTypes) and not getIsNumber(fillTypes) and not getIsString(fillTypes))
	or (getIsTable(fillTypes) and #fillTypes == 0)
	or (not getIsNumber(unitSet) and not getIsString(unitSet))
	then
		raiseError()
	end

	if not getIsTable(fillTypes) then fillTypes = ucMod:splitString(" ", fillTypes) end
	unitSet = ucMod:getUnitSet(unitSet)

	if unitSet.isFillTypeSet then
		local testUnit = ucMod.UNIT_UNKNOWN
		for _, fillType in pairs(fillTypes) do
			if ucMod:getIsFillTypeValid(fillType) then
				fillType = ucMod:getFillType(fillType)
				if testUnit == ucMod.UNIT_UNKNOWN then
					testUnit = fillType.units[unitSet.index]
				else
					if fillType.units[unitSet.index] ~= testUnit then
						testUnit = nil
						break
					end
				end
			end
		end

		if testUnit and testUnit ~= ucMod.UNIT_UNKNOWN then
			return ucMod:getUnit(testUnit)
		end
	end

	if unitSet.isFillTypeGroupSet then
		for _, fillTypeGroup in ipairs(ucMod.fillTypeGroupsByPriority) do
			for _, fillType in pairs(fillTypes) do
				if ucMod:getIsFillTypeValid(fillType) then
					fillType = ucMod:getFillType(fillType)
					if fillType.group == fillTypeGroup.index then
						return ucMod:getUnit(fillType.units[unitSet.index])
					end
				end
			end
		end
	end
	raiseError()
end

function UnitConvert:convertValue(value, unit, fillType)
-- value:		value to convert
-- unit:		unit used for conversion (by name or index)
-- fillType:	fillType used for weight conversions (must be "false" for non-fillType conversions)
-- return:		converted value

	if not getIsNumber(value) or (not getIsNumber(unit) and not getIsString(unit)) or fillType == nil then
		raiseError()
		return
	end

	unit = ucMod:getUnit(unit)
	if fillType and unit.unitType == ucMod.UNIT_TYPE_WEIGHT then
		fillType = ucMod:getFillType(fillType)
		value = value * fillType.massPerLiter * fillType.massFactor
	end
	value = value * unit.conversionFactor

	return value
end

function UnitConvert:getVehicleFillMass(vehicle)
-- vehicle:	vehicle used to calculate mass (table)
-- return:
--	[1]:	original fillType mass of vehicle
--	[2]:	converted fillType mass of vehicle

	if vehicle == nil or not getIsTable(vehicle) then
		raiseError()
		return
	end

	local massOld = 0
	local massNew = 0
    if vehicle.fillUnits then
        for _, vehicleFillUnit in pairs(vehicle.fillUnits) do
			local fillType = vehicleFillUnit.currentFillType
			if ucMod:getIsFillTypeValid(fillType) then
				local fillLevel = Utils.getNoNil(vehicleFillUnit.fillLevel, 0)
				fillType = ucMod:getFillType(fillType)
				massOld = massOld + (fillLevel * fillType.massPerLiter)
				massNew = massNew + (fillLevel * fillType.massPerLiter * fillType.massFactor)
            end
        end
    end
    return massOld, massNew
end

-- [HELPLINES] ****************************************************************************************************************************************************************************************

function UnitConvert:setupHelpLines()
	local helpLinesPath = ucMod.modDir.."helpLine/"
	local helpLinesFile = helpLinesPath.."helpLine.xml"
	if xAssert(fileExists(helpLinesFile)) then
		ucMod.helpLines = {}
		ucMod.helpLines.categories = {}

		local baseKey = "helpLines"
		local xmlFile = loadXMLFile(baseKey, helpLinesFile)

		if hasXMLProperty(xmlFile, baseKey) then
			local categoryIdx = 0
			while true do
				local categoryKey = string.format("%s.helpLineCategory(%d)", baseKey, categoryIdx)
				if not hasXMLProperty(xmlFile, categoryKey) then break end

				local category = {}
				category.helpLines = {}

				local categoryTitle = getXMLString(xmlFile, categoryKey.."#title")
				if xAssert(categoryTitle) then
					category.title = categoryTitle
				end

				g_inGameMenu.helpLineCategorySelectorElement:addText(getText(category.title))

				local lineIdx = 0
				while true do
					local lineKey = string.format("%s.helpLine(%d)", categoryKey, lineIdx)
					if not hasXMLProperty(xmlFile, lineKey) then break end

					local line = {}
					line.items = {}

					local lineTitle = getXMLString(xmlFile, lineKey.."#title")
					if xAssert(lineTitle) then
						line.title = lineTitle
					end

					local itemIdx = 0
					while true do
						local itemKey = string.format("%s.item(%d)", lineKey, itemIdx)
						if not hasXMLProperty(xmlFile, itemKey) then break end

						local item = {}

						local itemType			= getXMLString(xmlFile, itemKey.."#type")
						local itemValue			= getXMLString(xmlFile, itemKey.."#value")
						local itemHeightScale	= Utils.getNoNil(getXMLFloat(xmlFile, itemKey.."#heightScale"), 1.0)
						if xAssert(itemType and itemValue) then
							if itemType == "text" then
								item.value = itemValue
							elseif itemType == "image" then
								item.value = helpLinesPath..itemValue
							else
								raiseError()
							end
							item.type = itemType
							item.heightScale = itemHeightScale
						end

						table.insert(line.items, item)
						itemIdx = itemIdx + 1
					end
					table.insert(category.helpLines, line)
					lineIdx = lineIdx + 1
				end
				table.insert(ucMod.helpLines.categories, category)
				table.insert(g_inGameMenu.helpLineCategories, category)
				categoryIdx = categoryIdx + 1
			end
		end
	end
end

-- [CONFIGURATION MENU] *******************************************************************************************************************************************************************************

function UnitConvert:setupConfigMenu()
	local cfgMenu = ucMod.configMenu

	local function insertHeading(tbl, text)
	-- tbl:		headings table
	-- text:	text to insert

		if not getIsTable(tbl) or (not getIsNumber(text) and not getIsString(text)) then
			raiseError()
			return
		end

		local added = false
		text = tostring(text)
		for k, v in pairs(tbl) do
			if v == text then
				added = true
				break
			end
		end

		if not added then
			table.insert(tbl, text)
		end
	end

	for _, page in ipairs(cfgMenu.pages) do
		page.headings = {}
		page.listItems = {}

		if page.index == cfgMenu.PAGE_FILLTYPES then
			for _, fillType in ipairs(ucMod.fillTypesByIndex) do
				if fillType.showOnMenu then
					local fillTypeGroup = ucMod:getFillTypeGroup(fillType.group)

					local title = fillType.name
					local unitTypes = ucMod:getUnitTypesByGroup(fillTypeGroup.index)

					local columns = {}

					page.COLUMN_FILLTYPE_GROUP = 1
					local fillTypeGroupIdx = page.COLUMN_FILLTYPE_GROUP
					insertHeading(page.headings, getText("UC_menu_heading_group"))

					columns[fillTypeGroupIdx] = {}
					columns[fillTypeGroupIdx].valuesByIndex = {}
					table.insert(columns[fillTypeGroupIdx].valuesByIndex, {name = fillTypeGroup.name})
					columns[fillTypeGroupIdx].valueIndex = #columns[fillTypeGroupIdx].valuesByIndex
					columns[fillTypeGroupIdx].NUM_VALUES = #columns[fillTypeGroupIdx].valuesByIndex

					page.COLUMN_UNIT_SET_MIN = 2
					local unitSetIdx = page.COLUMN_UNIT_SET_MIN
					for _, unitSet in ipairs(ucMod.unitSetsByIndex) do
						if unitSet.isFillTypeSet then
							insertHeading(page.headings, unitSet.nameI18N)

							columns[unitSetIdx] = {}
							columns[unitSetIdx].valuesByIndex = {}
							if fillType.lockUnits then
								local unit = ucMod:getUnit(fillType.units[unitSet.index])
								table.insert(columns[unitSetIdx].valuesByIndex, {name = unit.name})
								columns[unitSetIdx].valueIndex = 1
							else
								for _, unit in ipairs(ucMod.unitsByIndex) do
									if unitTypes[unit.unitType] then
										table.insert(columns[unitSetIdx].valuesByIndex, {name = unit.name})

									end
									if fillType.units[unitSet.index] == unit.index then
										columns[unitSetIdx].valueIndex = #columns[unitSetIdx].valuesByIndex
									end
								end
							end
							columns[unitSetIdx].NUM_VALUES = #columns[unitSetIdx].valuesByIndex
							page.COLUMN_UNIT_SET_MAX = unitSetIdx
							unitSetIdx = unitSetIdx + 1
						end
					end

					page.COLUMN_MASS_FACTOR = page.COLUMN_UNIT_SET_MAX + 1
					local massFactorIdx = page.COLUMN_MASS_FACTOR
					insertHeading(page.headings, getText("UC_menu_heading_massFactor"))

					columns[massFactorIdx] = {}
					columns[massFactorIdx].valuesByIndex = {}
					for massFactor = ucMod.MASS_FACTOR_MIN, ucMod.MASS_FACTOR_MAX, ucMod.MASS_FACTOR_STEP do
						table.insert(columns[massFactorIdx].valuesByIndex, {name = tostring(math.floor(massFactor * 100)).."%"})
						if fillType.massFactor == massFactor then
							columns[massFactorIdx].valueIndex = #columns[massFactorIdx].valuesByIndex
						end
						columns[massFactorIdx].NUM_VALUES = #columns[massFactorIdx].valuesByIndex
					end

					table.insert(page.listItems, {title = title, columns = columns})
				end
			end
		elseif page.index == cfgMenu.PAGE_FILLTYPEGROUPS then
			for _, fillTypeGroup in ipairs(ucMod.fillTypeGroupsByPriority) do
				if fillTypeGroup.showOnMenu then
					local title = fillTypeGroup.name

					page.COLUMN_UNIT_SET_MIN = 1

					local columns = {}
					local unitSetIdx = 1
					for _, unitSet in ipairs(ucMod.unitSetsByIndex) do
						if unitSet.isFillTypeGroupSet then
							insertHeading(page.headings, unitSet.nameI18N)

							columns[unitSetIdx] = {}
							columns[unitSetIdx].valuesByIndex = {}
							if fillTypeGroup.lockUnits then
								local unit = ucMod:getUnit(fillTypeGroup.units[unitSetIdx])
								table.insert(columns[unitSetIdx].valuesByIndex, {name = unit.name})
								columns[unitSetIdx].valueIndex = 1
							else
								for _, unit in ipairs(ucMod.unitsByIndex) do
									if fillTypeGroup.unitTypes[unit.unitType] then
										table.insert(columns[unitSetIdx].valuesByIndex, {name = unit.name})
									end
									if fillTypeGroup.units[unitSet.index] == unit.index then
										columns[unitSetIdx].valueIndex = #columns[unitSetIdx].valuesByIndex
									end
								end
							end
							columns[unitSetIdx].NUM_VALUES = #columns[unitSetIdx].valuesByIndex
							page.COLUMN_UNIT_SET_MAX = unitSetIdx
							unitSetIdx = unitSetIdx + 1
						end
					end

					table.insert(page.listItems, {title = title, columns = columns})
				end
			end
		elseif page.index == cfgMenu.PAGE_OTHER then
			for _, unitType in ipairs(ucMod.unitTypesByIndex) do
				if unitType.index > ucMod.UNIT_TYPE_VOLUME then
					insertHeading(page.headings, getText("UC_menu_heading_unit"))

					local title = unitType.name

					local columns = {}
					columns[1] = {}
					columns[1].valuesByIndex = {}
					local column = columns[1]
					for _, unit in ipairs(ucMod.unitsByIndex) do
						if unit.unitType == unitType.index then
							table.insert(column.valuesByIndex, {name = unit.name})
						end
						if (unitType.index == ucMod.UNIT_TYPE_WEIGHT and unit.index == ucMod.weightUnit)
						or (unitType.index == ucMod.UNIT_TYPE_LENGTH and unit.index == ucMod.lengthUnit)
						or (unitType.index == ucMod.UNIT_TYPE_AREA and unit.index == ucMod.areaUnit)
						then
							column.valueIndex = #column.valuesByIndex
						end
					end
					column.NUM_VALUES = #column.valuesByIndex

					table.insert(page.listItems, {title = title, columns = columns})
				end
			end
		elseif page.index == cfgMenu.PAGE_UNITS then
			for _, unit in ipairs(ucMod.unitsByIndex) do
				insertHeading(page.headings, getText("UC_menu_heading_precision"))

				local title = unit.name

				local columns = {}
				columns[1] = {}
				columns[1].valuesByIndex = {}
				local column = columns[1]
				if unit.unitType == ucMod.UNIT_TYPE_PIECE
				or unit.unitType == ucMod.UNIT_TYPE_ANIMAL
				then
					table.insert(column.valuesByIndex, {name = "0"})
					column.valueIndex = 1
				else
					for idx = 0, ucMod.PRECISION_MAX do
						table.insert(column.valuesByIndex, {name = tostring(idx)})
						if unit.precision == idx then
							column.valueIndex = #column.valuesByIndex
						end
					end
				end
				column.NUM_VALUES = #column.valuesByIndex

				table.insert(page.listItems, {title = title, columns = columns})
			end
		else
			raiseError()
			return
		end

		page.NUM_COLUMNS = #page.headings
		if page.NUM_COLUMNS > 0 then page.columnIndex = 1 else page.columnIndex = 0 end

		page.NUM_ROWS = #page.listItems
		if page.NUM_ROWS > 0 then
			page.listIndex = 1
			page.startRow = 1
		else
			page.listIndex = 0
			page.startRow = 0
		end
	end

	setTextBold(true)
	for _, page in pairs(cfgMenu.pages) do
		page.textHeight = getTextHeight(cfgMenu.textSize, page.nameI18N)
		page.titleTextWidth = getTextWidth(cfgMenu.textSize, page.nameI18N)
		page.columnTextWidth = {}

		for _, listItem in pairs(page.listItems) do
			local tWidth = getTextWidth(cfgMenu.textSize, listItem.title)
			local tHeight = getTextHeight(cfgMenu.textSize, listItem.title)

			page.titleTextWidth = math.max(page.titleTextWidth, tWidth)
			page.textHeight = math.max(page.textHeight, tHeight)

			for columnIdx, column in pairs(listItem.columns) do
				column.valuesByName = Utils.getNoNil(column.valuesByName, {})

				tWidth = getTextWidth(cfgMenu.textSize, page.headings[columnIdx])
				tHeight = getTextHeight(cfgMenu.textSize, page.headings[columnIdx])

				page.columnTextWidth[columnIdx] = Utils.getNoNil(page.columnTextWidth[columnIdx], 0)
				page.columnTextWidth[columnIdx] = math.max(page.columnTextWidth[columnIdx], tWidth)
				page.textHeight = math.max(page.textHeight, tHeight)

				for valueIdx, value in ipairs(column.valuesByIndex) do
					value.index = valueIdx
					column.valuesByName[value.name] = value

					tWidth = getTextWidth(cfgMenu.textSize, value.name)
					tHeight = getTextHeight(cfgMenu.textSize, value.name)

					page.columnTextWidth[columnIdx] = math.max(page.columnTextWidth[columnIdx], tWidth)
					page.textHeight = math.max(page.textHeight, tHeight)
				end
			end
		end

		page.textHeight = page.textHeight - (0.0025 * uiScale) -- Adjustment
		page.overlayWidth = page.titleTextWidth + (cfgMenu.marginX * 2)

		for _, textWidth in pairs(page.columnTextWidth) do
			page.overlayWidth = page.overlayWidth + (cfgMenu.lineSpacingX + textWidth)
		end
	end
	setTextBold(false)
end

function UnitConvert:drawConfigMenu()
	local cfgMenu = ucMod.configMenu
	local cfgMenuPage = cfgMenu.pages[cfgMenu.pageIndex]

	setTextAlignment(RenderText.ALIGN_LEFT)
	setTextColor(unpack(cfgMenu.colors.titleText))
	setTextBold(true)

	local cfgMenuWidth = cfgMenuPage.overlayWidth
	local cfgMenuHeight = cfgMenuPage.textHeight + (cfgMenu.marginY * 2)

	local cfgMenuX = g_currentMission.infoBarBgOverlay.x + g_currentMission.infoBarBgOverlay.width - cfgMenuWidth
	local cfgMenuY = g_currentMission.infoBarBgOverlay.y - cfgMenuHeight - cfgMenu.marginY

	if g_seasons and g_seasons.forecast.visible then
		cfgMenuY = g_seasons.forecast.forecastY - cfgMenuHeight - g_seasons.forecast.forecastSpacingHeight
	end

	setOverlayColor(cfgMenu.overlayId, unpack(cfgMenu.colors.titleBg))
	renderOverlay(cfgMenu.overlayId, cfgMenuX, cfgMenuY, cfgMenuWidth, cfgMenuHeight)

	local cfgMenuTextX = cfgMenuX + cfgMenu.marginX
	local cfgMenuTextY = cfgMenuY + cfgMenu.marginY

	renderText(cfgMenuTextX, cfgMenuTextY, cfgMenu.textSize, cfgMenuPage.nameI18N)

	local currentColTextX = cfgMenuTextX + cfgMenuPage.titleTextWidth + cfgMenu.lineSpacingX
	local textOffsetX = currentColTextX

	setTextAlignment(RenderText.ALIGN_CENTER)
	for columnIdx = 1, cfgMenuPage.NUM_COLUMNS do
		textOffsetX = textOffsetX + (cfgMenuPage.columnTextWidth[columnIdx] * 0.5)
		renderText(textOffsetX, cfgMenuTextY, cfgMenu.textSize, cfgMenuPage.headings[columnIdx])
		textOffsetX = textOffsetX + (cfgMenuPage.columnTextWidth[columnIdx] * 0.5) + cfgMenu.lineSpacingX
	end
	setTextAlignment(RenderText.ALIGN_LEFT)

	cfgMenuHeight = cfgMenu.marginY
	cfgMenuY = cfgMenuY - cfgMenuHeight
	setOverlayColor(cfgMenu.overlayId, unpack(cfgMenu.colors.itemBg))
	renderOverlay(cfgMenu.overlayId, cfgMenuX, cfgMenuY, cfgMenuWidth, cfgMenuHeight)

	local function addListItem(idx, isSelected)
	-- idx:			index of menu item
	-- isSelected:	Determines whether menu item is selected or not

		if (idx and not getIsNumber(idx)) then
			raiseError()
			return
		end

		local itemPosX = cfgMenuX + cfgMenu.marginX
		local itemWidth = cfgMenuWidth - (cfgMenu.marginX * 2)
		cfgMenuY = cfgMenuY - cfgMenuHeight

		setOverlayColor(cfgMenu.overlayId, unpack(cfgMenu.colors.itemBg))
		renderOverlay(cfgMenu.overlayId, cfgMenuX, cfgMenuY, cfgMenuWidth, cfgMenuHeight)

		if isSelected then
			setOverlayColor(cfgMenu.overlayId, unpack(cfgMenu.colors.itemBgSelected))
			setTextColor(unpack(cfgMenu.colors.itemTextSelected))
			setTextBold(true)
		else
			if not idx then
				setOverlayColor(cfgMenu.overlayId, unpack(cfgMenu.colors.itemBg))
			elseif idx % 2 == 0 then
				setOverlayColor(cfgMenu.overlayId, unpack(cfgMenu.colors.itemBgEven))
			else
				setOverlayColor(cfgMenu.overlayId, unpack(cfgMenu.colors.itemBgOdd))
			end
			setTextColor(unpack(cfgMenu.colors.itemText))
			setTextBold(false)
		end

		renderOverlay(cfgMenu.overlayId, itemPosX, cfgMenuY, itemWidth, cfgMenuHeight)
	end

	local minRow = cfgMenuPage.startRow
	local maxRow = minRow + (cfgMenu.NUM_VISIBLE_ROWS - 1)

	cfgMenuHeight = cfgMenuPage.textHeight + (cfgMenu.lineSpacingY * 2)

	for listIdx, listItem in ipairs(cfgMenuPage.listItems) do
		if listIdx == (minRow - 1) or listIdx == (maxRow + 1) then
			local arrowColor = cfgMenu.colors.itemTextSelected
			local arrowWidth = cfgMenu.marginX * 2
			local arrowHeight = cfgMenu.marginY
			local arrowX = (cfgMenuX + (cfgMenuWidth / 2)) - (arrowWidth / 2)
			local arrowY = cfgMenuY

			if listIdx == (minRow - 1) then
				setOverlayColor(cfgMenu.arrowUpId, unpack(arrowColor))
				renderOverlay(cfgMenu.arrowUpId, arrowX, arrowY, arrowWidth, arrowHeight)
			else
				setOverlayColor(cfgMenu.arrowDownId, unpack(arrowColor))
				renderOverlay(cfgMenu.arrowDownId, arrowX, arrowY, arrowWidth, arrowHeight)
			end
		elseif listIdx >= minRow and listIdx <= maxRow then
			local isSelected = false
			if listIdx == cfgMenuPage.listIndex then isSelected = true end

			addListItem(listIdx, isSelected)
			cfgMenuTextY = cfgMenuY + cfgMenu.lineSpacingY
			renderText(cfgMenuTextX, cfgMenuTextY, cfgMenu.textSize, listItem.title)

			textOffsetX = currentColTextX

			setTextAlignment(RenderText.ALIGN_CENTER)
			for columnIdx, column in ipairs(listItem.columns) do
				if listIdx == cfgMenuPage.listIndex and columnIdx == cfgMenuPage.columnIndex then
					setTextColor(unpack(cfgMenu.colors.itemTextSelected))
					setTextBold(true)
				else
					setTextColor(unpack(cfgMenu.colors.itemText))
					setTextBold(false)
				end
				textOffsetX = textOffsetX + (cfgMenuPage.columnTextWidth[columnIdx] * 0.5)
				renderText(textOffsetX, cfgMenuTextY, cfgMenu.textSize, column.valuesByIndex[column.valueIndex].name)
				textOffsetX = textOffsetX + (cfgMenuPage.columnTextWidth[columnIdx] * 0.5) + cfgMenu.lineSpacingX
			end
			setTextAlignment(RenderText.ALIGN_LEFT)

			if listIdx == cfgMenuPage.NUM_ROWS or listIdx == maxRow then
				cfgMenuHeight = cfgMenu.marginY
				cfgMenuY = cfgMenuY - cfgMenuHeight
				setOverlayColor(cfgMenu.overlayId, unpack(cfgMenu.colors.itemBg))
				renderOverlay(cfgMenu.overlayId, cfgMenuX, cfgMenuY, cfgMenuWidth, cfgMenuHeight)
			end
		end
	end

	setTextAlignment(RenderText.ALIGN_LEFT)
	setTextColor(1, 1, 1, 1)
	setTextBold(false)
end

-- [THIRD PARTY ADDONS] *******************************************************************************************************************************************************************************

function UnitConvert:loadThirdPartyAddOns()
	for idx, addOn in ipairs(ucMod.addOns) do
		local source = addOn.source
		local name = Utils.getNoNil(addOn.name, source)

		for modName, isValid in pairs(g_modIsLoaded) do
			if isValid then
				local modDesc = ModsUtil.modNameToMod[modName]
				if modDesc and modDesc.version then
					local mod = getfenv(0)[modName]
					if mod[source] then
						if modDesc.version == addOn.version then
							addOn.target = mod[source]
							if addOn:initialize() then
								addOn.index = idx
								debugMsg(-1, "UC_info_addOnLoaded", name, modDesc.version)
							else
								debugMsg(-1, "UC_error_addOnLoadFailed", name, modDesc.version)
							end
						else
							debugMsg(-1, "UC_error_addOnVersionInvalid", name, modDesc.version, addOn.version)
						end
					end
				end
			end
		end
	end
end

-- [CONFIGURATION FILES] ******************************************************************************************************************************************************************************

function UnitConvert:getModI18N(modName)
	modName = getfenv(0)[modName]
	if xAssert(modName and modName.g_i18n) then
		for name, _ in pairs(modName.g_i18n.texts) do
			if string.find(name, "^UC_") then
				if not g_i18n:hasText(name) then
					g_i18n:setText(name, modName.g_i18n:getText(name))
				end
			end
		end
	end
end

function UnitConvert:addConfigXML(fileName, allowNew, allowOverwrite)
-- fileName:		fileName to add
-- allowNew:		allow this file to create new values
-- allowOverwrite:	allow this file to overwrite values from previous files
-- return:			success

	if not getIsString(fileName)
	or (allowNew and not getIsBoolean(allowNew))
	or (allowOverwrite and not getIsBoolean(allowOverwrite))
	then
		raiseError()
		return
	end

	if fileExists(fileName) then
		table.insert(ucMod.configXMLs, {fileName = fileName, allowNew = allowNew, allowOverwrite = allowOverwrite})
		return true
	end
end

function UnitConvert:loadConfig()
	local configFileName = "unitConvert"
	ucMod.configXMLs = {}

	-- Add default config(s)
	xAssert(ucMod:addConfigXML(ucMod.modDir.."config/"..configFileName..".xml", true, true))
	ucMod:addConfigXML(ucMod.modDir.."config/"..configFileName..g_languageSuffix..".xml", true, true)

	-- Add mod map (if applicable)
	local mapName = g_modSelectionScreen.mapModName
	if mapName then
		local mapFile = g_modNameToDirectory[mapName]..configFileName..".xml"
		if ucMod:addConfigXML(mapFile, true, true) then
			ucMod:getModI18N(mapName)
		end
	end

	-- Add additional loaded mods (if any)
	for modName, isValid in pairs(g_modIsLoaded) do
		if isValid then
			local modFile = g_modNameToDirectory[modName]..configFileName..".xml"
			if modName ~= mapName and modName ~= g_i18n.modName and ucMod:addConfigXML(modFile, true, false) then
				ucMod:getModI18N(modName)
			end
		end
	end

	-- Add saved config (if any)
	local profilePath = getUserProfileAppPath()
	ucMod.saveFile = profilePath..configFileName..".xml"
	ucMod:addConfigXML(ucMod.saveFile, false, true)

	if #ucMod.configXMLs == 0 then
		raiseError()
		return
	end

	-- Load configuration data
	ucMod.unitsByName = {}
	ucMod.unitsByIndex = {}

	ucMod.fillTypeGroupsByName = {}
	ucMod.fillTypeGroupsByIndex = {}

	ucMod.fillTypesByName = {}
	ucMod.fillTypesByIndex = {}

	ucMod.yieldDisplay = ucMod.YIELD_VEHICLE

	local function getDefaultFillTypeGroup(fillType)
	-- fillType: 	fillType to query (by index or name)
	-- return:		default fillType group for selected fillType (index)

		fillType = xAssert(Utils.getNoNil(tonumber(fillType), fillType))
		fillType = xAssert(Utils.getNoNil(FillUtil.fillTypeNameToDesc[fillType], FillUtil.fillTypeIndexToDesc[fillType]))

		if FillUtil.categoryToFillTypes[FillUtil.FILLTYPE_CATEGORY_COMBINE][fillType.index] then
			return ucMod.FILLTYPE_GROUP_GRAIN
		elseif FillUtil.categoryToFillTypes[FillUtil.FILLTYPE_CATEGORY_WINDROW][fillType.index]
		or FillUtil.categoryToFillTypes[FillUtil.FILLTYPE_CATEGORY_FORAGEWAGON][fillType.index]
		then
			return ucMod.FILLTYPE_GROUP_WINDROW
		elseif FillUtil.categoryToFillTypes[FillUtil.FILLTYPE_CATEGORY_SPREADER][fillType.index]
		or FillUtil.categoryToFillTypes[FillUtil.FILLTYPE_CATEGORY_MANURESPREADER][fillType.index]
		then
			return ucMod.FILLTYPE_GROUP_FERTILIZER
		elseif FillUtil.categoryToFillTypes[FillUtil.FILLTYPE_CATEGORY_LIQUID][fillType.index]
		or FillUtil.categoryToFillTypes[FillUtil.FILLTYPE_CATEGORY_SPRAYER][fillType.index]
		or FillUtil.categoryToFillTypes[FillUtil.FILLTYPE_CATEGORY_SLURRYTANK][fillType.index]
		then
			return ucMod.FILLTYPE_GROUP_LIQUID
		elseif FillUtil.categoryToFillTypes[FillUtil.FILLTYPE_CATEGORY_PIECE][fillType.index] then
			return ucMod.FILLTYPE_GROUP_PIECE
		elseif FillUtil.categoryToFillTypes[FillUtil.FILLTYPE_CATEGORY_ANIMAL][fillType.index] then
			return ucMod.FILLTYPE_GROUP_ANIMAL
		else
			local fruitType = FruitUtil.fillTypeToFruitType[fillType.index]
			if fruitType then
				local fruitDesc = FruitUtil.fruitIndexToDesc[fruitType]
				if fruitDesc.preparedGrowthState and fruitDesc.preparedGrowthState > 0 then
					return ucMod.FILLTYPE_GROUP_ROOTCROP
				end
			end
			return ucMod.FILLTYPE_GROUP_BULK
		end
	end

	local function getDefaultUnitByTypes(unitTypes)
	-- unitTypes:	Unit types to query(by name, index, or table)
	-- return:		Default unit for selected unit types (index)

		if not getIsTable(unitTypes) and not getIsNumber(unitTypes) and not getIsString(unitTypes) then
			raiseError()
			return
		end

		if not getIsTable(unitTypes) then
			local unitTypeList = ucMod:splitString(" ", unitTypes)
			if xAssert(#unitTypeList > 0) then
				unitTypes = {}
				for _, unitType in pairs(unitTypeList) do
					unitTypes[unitType] = true
				end
			end
		end

		if unitTypes[ucMod.UNIT_TYPE_VOLUME] then
			return ucMod.UNIT_LITER
		elseif unitTypes[ucMod.UNIT_TYPE_WEIGHT] then
			return ucMod.UNIT_METRICTON
		elseif unitTypes[ucMod.UNIT_TYPE_PIECE] then
			return ucMod.UNIT_PIECE
		elseif unitTypes[ucMod.UNIT_TYPE_ANIMAL] then
			return ucMod.UNIT_ANIMAL
		end

		raiseError()
	end

	for configIdx, configData in ipairs(ucMod.configXMLs) do
		local fileName = configData.fileName
		local allowNew = configData.allowNew
		local allowOverwrite = configData.allowOverwrite

		local baseKey = "UnitConvert"
		local configXML = loadXMLFile(baseKey, fileName)
		if hasXMLProperty(configXML, baseKey) then

			local skipConfig = false

			local configVersion = Utils.getNoNil(getXMLInt(configXML, baseKey.."#version"), 0)
			if configIdx == 1 then
				ucMod.configVersion = configVersion
			else
				if configVersion < ucMod.configVersion then
					debugMsg(-1, tostring(fileName))
					debugMsg(-1, "UC_warning_configIsOld")
					skipConfig = true
				end
			end

			if not skipConfig then
				if fileName == ucMod.saveFile then
					local yieldDisplay = getXMLString(configXML, baseKey.."#yieldDisplay")
					if yieldDisplay ~= ucMod.YIELD_VEHICLE and yieldDisplay ~= ucMod.YIELD_PRICE and yieldDisplay ~= ucMod.YIELD_OFF then
						yieldDisplay = nil
					end
					ucMod.yieldDisplay = Utils.getNoNil(yieldDisplay, ucMod.yieldDisplay)
				end

				local idx = 0
				while true do
					local unitKey = string.format("%s.unit(%d)", baseKey, idx)
					if not hasXMLProperty(configXML, unitKey) then
						break
					end

					local unitName = getXMLString(configXML, unitKey.."#name")
					if not unitName then
						debugMsg(-1, tostring(fileName))
						debugMsg(-1, "UC_error_unitInvalid")
					else
						local unitType				= getXMLString(configXML, unitKey.."#type")
						local unitPrecision			= getXMLInt(configXML, unitKey.."#precision")
						local unitConversionFactor	= getXMLFloat(configXML, unitKey.."#conversionFactor")
						local unitNameI18N			= getText("UC_unit_"..unitName)
						local unitNameI18NShort		= getText("UC_unit_"..unitName.."Short")

						if unitType then
							unitType = ucMod.unitTypesByName[unitType]
							if not unitType then
								debugMsg(-1, tostring(fileName))
								debugMsg(-1, "UC_warning_unitInvalidType", unitName)
								unitType = nil
							else
								unitType = unitType.index
							end
						end

						if unitPrecision then
							if unitPrecision < 0 or unitPrecision > ucMod.PRECISION_MAX then
								debugMsg(-1, tostring(fileName))
								debugMsg(-1, "UC_warning_unitInvalidPrecision", unitName)
								unitPrecision = nil
							else
								if unitType == ucMod.UNIT_TYPE_PIECE or unitType == ucMod.UNIT_TYPE_ANIMAL then
									unitPrecision = nil
								end
							end
						end

						if unitConversionFactor then
							if unitConversionFactor <= 0 then
								debugMsg(-1, tostring(fileName))
								debugMsg(-1, "UC_warning_unitInvalidConversionFactor", unitName)
								unitConversionFactor = nil
							else
								if unitType == ucMod.UNIT_TYPE_PIECE or unitType == ucMod.UNIT_TYPE_ANIMAL then
									unitConversionFactor = nil
								end
							end
						end

						local unit = ucMod.unitsByName[unitName]
						if unit then
							if allowOverwrite then
								unit.precision			= Utils.getNoNil(unitPrecision, unit.precision)
								unit.conversionFactor	= Utils.getNoNil(unitConversionFactor, unit.conversionFactor)
							end
						else
							if allowNew then
								local unitIndex = #ucMod.unitsByIndex + 1
								unit = {
									name				= unitName,
									index				= unitIndex,
									unitType			= Utils.getNoNil(unitType, ucMod.UNIT_TYPE_PIECE),
									precision			= Utils.getNoNil(unitPrecision, 0),
									conversionFactor	= Utils.getNoNil(unitConversionFactor, 1),
									nameI18N			= unitNameI18N,
									nameI18NShort		= unitNameI18NShort}

								table.insert(ucMod.unitsByIndex, unit)
								ucMod.unitsByName[unit.name] = unit
								ucMod["UNIT_"..string.upper(unit.name)] = unit.index
								debugMsg(1, "UC_info_unitRegistered", unit.name)
							end
						end
					end
					idx = idx + 1
				end

				idx = 0
				while true do
					local fillTypeGroupKey = string.format("%s.fillTypeGroup(%d)", baseKey, idx)
					if not hasXMLProperty(configXML, fillTypeGroupKey) then
						break
					end

					local fillTypeGroupName = getXMLString(configXML, fillTypeGroupKey.."#name")
					if not fillTypeGroupName then
						debugMsg(-1, tostring(fileName))
						debugMsg(-1, "UC_error_fillTypeGroupInvalid")
					else
						local fillTypeGroupPriority		= getXMLInt(configXML, fillTypeGroupKey.."#priority")
						local fillTypeGroupUnits		= {}
						local fillTypeGroupUnitTypes	= getXMLString(configXML, fillTypeGroupKey.."#unitTypes")
						local fillTypeGroupLockUnits	= getXMLBool(configXML, fillTypeGroupKey.."#lockUnits")
						local fillTypeGroupShowOnMenu	= getXMLBool(configXML, fillTypeGroupKey.."#showOnMenu")
						local fillTypeGroupNameI18N 	= getText("UC_fillType_group_"..fillTypeGroupName)

						if fillTypeGroupPriority then
							if fillTypeGroupPriority <= 0 then
								debugMsg(-1, tostring(fileName))
								debugMsg(-1, "UC_warning_fillTypeGroupInvalidPriority", fillTypeGroupName)
								fillTypeGroupPriority = nil
							end
						end

						if fillTypeGroupUnitTypes then
							local unitTypes = ucMod:splitString(" ", fillTypeGroupUnitTypes)
							fillTypeGroupUnitTypes = {}
							for _, unitType in pairs(unitTypes) do
								if not ucMod.unitTypesByName[unitType] then
									debugMsg(-1, tostring(fileName))
									debugMsg(-1, "UC_warning_fillTypeGroupInvalidUnitType", fillTypeGroupName)
								else
									fillTypeGroupUnitTypes[ucMod.unitTypesByName[unitType].index] = true
								end
							end
						end
						if fillTypeUnitGroupTypes == {} then fillTypeUnitGroupTypes = nil end

						for _, unitSet in ipairs(ucMod.unitSetsByIndex) do
							if unitSet.isFillTypeGroupSet then
								fillTypeGroupUnits[unitSet.index] = getXMLString(configXML, fillTypeGroupKey.."#unit"..tostring(unitSet.index))
								if fillTypeGroupUnits[unitSet.index] then
									local unit = ucMod.unitsByName[fillTypeGroupUnits[unitSet.index]]
									if not unit
									or (unit.unitType ~= ucMod.UNIT_TYPE_VOLUME
									and unit.unitType ~= ucMod.UNIT_TYPE_WEIGHT
									and unit.unitType ~= ucMod.UNIT_TYPE_PIECE
									and unit.unitType ~= ucMod.UNIT_TYPE_ANIMAL)
									then
										debugMsg(-1, tostring(fileName))
										debugMsg(-1, "UC_warning_fillTypeGroupInvalidUnit", fillTypeGroupName, "unit"..tostring(unitSet.index))
										fillTypeGroupUnits[unitSet.index] = nil
									else
										fillTypeGroupUnits[unitSet.index] = unit.index
									end
								end
							end
						end

						local fillTypeGroup = ucMod.fillTypeGroupsByName[fillTypeGroupName]
						if fillTypeGroup then
							if allowOverwrite then
								fillTypeGroup.priority		= Utils.getNoNil(fillTypeGroupPriority, fillTypeGroup.priority)
								fillTypeGroup.lockUnits		= Utils.getNoNil(fillTypeGroupLockUnits, fillTypeGroup.lockUnits)
								fillTypeGroup.showOnMenu	= Utils.getNoNil(fillTypeGroupShowOnMenu, fillTypeGroup.showOnMenu)

								if not fillTypeGroup.lockUnits then
									for unitIdx, _ in pairs(fillTypeGroup.units) do
										if fillTypeGroupUnits[unitIdx] then
											local unit = ucMod.unitsByIndex[fillTypeGroupUnits[unitIdx]]
											if fillTypeGroup.unitTypes[unit.unitType] then
												fillTypeGroup.units[unitIdx] = fillTypeGroupUnits[unitIdx]
											end
										end
									end
								end
							end
						else
							if allowNew then
								local fillTypeGroupIndex = #ucMod.fillTypeGroupsByIndex + 1
								fillTypeGroup = {
									name		= fillTypeGroupName,
									index		= fillTypeGroupIndex,
									priority	= Utils.getNoNil(fillTypeGroupPriority, fillTypeGroupIndex),
									units 		= {},
									unitTypes	= Utils.getNoNil(fillTypeGroupUnitTypes, {
										[ucMod.UNIT_TYPE_VOLUME]	= true,
										[ucMod.UNIT_TYPE_WEIGHT]	= true}),
									lockUnits	= Utils.getNoNil(fillTypeGroupLockUnits, false),
									showOnMenu	= Utils.getNoNil(fillTypeGroupShowOnMenu, true),
									nameI18N	= fillTypeGroupNameI18N}

								for _, unitSet in pairs(ucMod.unitSetsByIndex) do
									if unitSet.isFillTypeGroupSet then
										if fillTypeGroupUnits[unitSet.index] then
											local unit = ucMod.unitsByIndex[fillTypeGroupUnits[unitSet.index]]
											if not fillTypeGroup.unitTypes[unit.unitType] then
												fillTypeGroupUnits[unitSet.index] = nil
											end
										end
										fillTypeGroup.units[unitSet.index] = Utils.getNoNil(fillTypeGroupUnits[unitSet.index], getDefaultUnitByTypes(fillTypeGroup.unitTypes))
									end
								end

								table.insert(ucMod.fillTypeGroupsByIndex, fillTypeGroup)
								ucMod.fillTypeGroupsByName[fillTypeGroup.name] = fillTypeGroup
								ucMod["FILLTYPE_GROUP_"..string.upper(fillTypeGroup.name)] = fillTypeGroup.index
								debugMsg(1, "UC_info_fillTypeGroupRegistered", fillTypeGroup.name)
							end
						end
					end
					idx = idx + 1
				end

				idx = 0
				while true do
					local fillTypeKey = string.format("%s.fillType(%d)", baseKey, idx)
					if not hasXMLProperty(configXML, fillTypeKey) then
						break
					end

					local fillTypeName = getXMLString(configXML, fillTypeKey.."#name")
					if not fillTypeName then
						debugMsg(-1, tostring(fileName))
						debugMsg(-1, "UC_error_fillTypeInvalid")
					elseif not FillUtil.fillTypeNameToInt[fillTypeName] then
						debugMsg(10, tostring(fileName))
						debugMsg(10, "UC_info_fillTypeMissing", fillTypeName)
					else
						local fillTypeIndex				= FillUtil.fillTypeNameToInt[fillTypeName]
						local fillTypeGroup				= getXMLString(configXML, fillTypeKey.."#group")
						local fillTypeUnits				= {}
						local fillTypeLockUnits			= getXMLBool(configXML, fillTypeKey.."#lockUnits")
						local fillTypeShowOnMenu		= getXMLBool(configXML, fillTypeKey.."#showOnMenu")
						local fillTypeShowOnPriceTable	= getXMLBool(configXML, fillTypeKey.."#showOnPriceTable")
						local fillTypeMassFactor		= getXMLFloat(configXML, fillTypeKey.."#massFactor")

						if fillTypeGroup then
							if not ucMod.fillTypeGroupsByName[fillTypeGroup] then
								debugMsg(-1, tostring(fileName))
								debugMsg(-1, "UC_warning_fillTypeInvalidGroup", fillTypeName)
								fillTypeGroup = nil
							else
								fillTypeGroup = ucMod.fillTypeGroupsByName[fillTypeGroup].index
							end
						end

						for _, unitSet in pairs(ucMod.unitSetsByIndex) do
							if unitSet.isFillTypeSet then
								fillTypeUnits[unitSet.index] = getXMLString(configXML, fillTypeKey.."#unit"..tostring(unitSet.index))
								if fillTypeUnits[unitSet.index] then
									local unit = ucMod.unitsByName[fillTypeUnits[unitSet.index]]
									if not unit
									or (unit.unitType ~= ucMod.UNIT_TYPE_VOLUME
									and unit.unitType ~= ucMod.UNIT_TYPE_WEIGHT
									and unit.unitType ~= ucMod.UNIT_TYPE_PIECE
									and unit.unitType ~= ucMod.UNIT_TYPE_ANIMAL)
									then
										debugMsg(-1, tostring(fileName))
										debugMsg(-1, "UC_warning_fillTypeInvalidUnit", fillTypeName, "unit"..tostring(unitSet.index))
										fillTypeUnits[unitSet.index] = nil
									else
										if ucMod.unitsByName[fillTypeUnits[unitSet.index]].unitType == ucMod.UNIT_TYPE_WEIGHT
										and FillUtil.fillTypeNameToDesc[fillTypeName].massPerLiter >= 0.010000 then
											debugMsg(-1, "UC_warning_fillTypeHighMass", fillTypeName)
										end
										fillTypeUnits[unitSet.index] = unit.index
									end
								end
							end
						end

						if fillTypeMassFactor then
							if fillTypeMassFactor <= 0 then
								debugMsg(-1, tostring(fileName))
								debugMsg(-1, "UC_warning_fillTypeInvalidMassFactor")
								fillTypeMassFactor = nil
							end
						end

						local fillType = ucMod.fillTypesByName[fillTypeName]
						if fillType then
							if allowOverwrite then
								fillType.group				= Utils.getNoNil(fillTypeGroup, fillType.group)
								fillType.lockUnits			= Utils.getNoNil(fillTypeLockUnits, fillType.lockUnits)
								fillType.showOnMenu			= Utils.getNoNil(fillTypeShowOnMenu, fillType.showOnMenu)
								fillType.showOnPriceTable	= Utils.getNoNil(fillTypeShowOnPriceTable, fillType.showOnPriceTable)
								fillType.massFactor			= Utils.getNoNil(fillTypeMassFactor, fillType.massFactor)

								if not fillType.lockUnits then
									local unitTypes = ucMod:getUnitTypesByGroup(fillType.group)
									for unitIdx, _ in pairs(fillType.units) do
										if fillTypeUnits[unitIdx] then
											local unit = ucMod.unitsByIndex[fillTypeUnits[unitIdx]]
											if unitTypes[unit.unitType] then
												fillType.units[unitIdx] = fillTypeUnits[unitIdx]
											end
										end
									end
								end
							end
						else
							fillType = {
								name				= fillTypeName,
								index				= fillTypeIndex,
								group				= Utils.getNoNil(fillTypeGroup, getDefaultFillTypeGroup(fillTypeName)),
								units				= {},
								lockUnits			= Utils.getNoNil(fillTypeLockUnits, false),
								showOnMenu			= Utils.getNoNil(fillTypeShowOnMenu, true),
								showOnPriceTable	= Utils.getNoNil(fillTypeShowOnPriceTable, FillUtil.fillTypeIndexToDesc[fillTypeIndex].showOnPriceTable),
								massPerLiter		= FillUtil.fillTypeIndexToDesc[fillTypeIndex].massPerLiter,
								massFactor			= Utils.getNoNil(fillTypeMassFactor, 1)}

							local unitTypes = ucMod:getUnitTypesByGroup(fillType.group)
							for _, unitSet in ipairs(ucMod.unitSetsByIndex) do
								if unitSet.isFillTypeSet then
									if fillTypeUnits[unitSet.index] then
										local unit = ucMod.unitsByIndex[fillTypeUnits[unitSet.index]]
										if not unitTypes[unit.unitType] then
											fillTypeUnits[unitSet.index] = nil
										end
									end
									fillType.units[unitSet.index] = Utils.getNoNil(fillTypeUnits[unitSet.index], getDefaultUnitByTypes(unitTypes))
								end
							end

							ucMod.fillTypesByIndex[fillType.index] = fillType
							ucMod.fillTypesByName[fillType.name] = fillType
							debugMsg(1, "UC_info_fillTypeRegistered", fillType.name)
						end
					end
					idx = idx + 1
				end

				if allowOverwrite then
					local lengthUnit = getXMLString(configXML, baseKey.."#lengthUnit")
					if lengthUnit then
						if not ucMod.unitsByName[lengthUnit]
						or ucMod.unitsByName[lengthUnit].unitType ~= ucMod.UNIT_TYPE_LENGTH
						then
							debugMsg(-1, tostring(fileName))
							debugMsg(-1, "UC_warning_lengthUnitInvalid", lengthUnit)
							lengthUnit = nil
						else
							lengthUnit = ucMod.unitsByName[lengthUnit].index
						end
					end
					ucMod.lengthUnit = Utils.getNoNil(lengthUnit, Utils.getNoNil(ucMod.lengthUnit, ucMod.UNIT_METER))

					local weightUnit = getXMLString(configXML, baseKey.."#weightUnit")
					if weightUnit then
						if not ucMod.unitsByName[weightUnit]
						or ucMod.unitsByName[weightUnit].unitType ~= ucMod.UNIT_TYPE_WEIGHT
						then
							debugMsg(-1, tostring(fileName))
							debugMsg(-1, "UC_warning_weightUnitInvalid", weightUnit)
							weightUnit = nil
						else
							weightUnit = ucMod.unitsByName[weightUnit].index
						end
					end
					ucMod.weightUnit = Utils.getNoNil(weightUnit, Utils.getNoNil(ucMod.weightUnit, ucMod.UNIT_KILOGRAM))

					local areaUnit = getXMLString(configXML, baseKey.."#areaUnit")
					if areaUnit then
						if not ucMod.unitsByName[areaUnit]
						or ucMod.unitsByName[areaUnit].unitType ~= ucMod.UNIT_TYPE_AREA
						then
							debugMsg(-1, tostring(fileName))
							debugMsg(-1, "UC_warning_areaUnitInvalid", areaUnit)
							areaUnit = nil
						else
							areaUnit = ucMod.unitsByName[areaUnit].index
						end
					end
					ucMod.areaUnit = Utils.getNoNil(areaUnit, Utils.getNoNil(ucMod.areaUnit, ucMod.UNIT_HECTARE))
				end
			end
		end
		delete(configXML)
	end

	-- Finalize unit registration
	ucMod.UNIT_UNKNOWN = 0
	ucMod.NUM_UNITS = #ucMod.unitsByIndex

	-- Finalize fill type group registration
	ucMod.FILLTYPE_GROUP_UNKNOWN = 0
	ucMod.NUM_FILLTYPE_GROUPS = #ucMod.fillTypeGroupsByIndex
	ucMod.fillTypeGroupsByPriority = sortTable(ucMod.fillTypeGroupsByIndex, "priority")

	-- Finalize fill type registration
	for _, fillType in pairs(FillUtil.fillTypeIndexToDesc) do
		if fillType.index ~= FillUtil.FILLTYPE_UNKNOWN then
			local ucFillType = ucMod.fillTypesByIndex[fillType.index]
			if not ucFillType then
				ucFillType = {
					name				= fillType.name,
					index				= fillType.index,
					group				= getDefaultFillTypeGroup(fillType.name),
					units				= {},
					lockUnits			= false,
					showOnMenu			= true,
					showOnPriceTable	= fillType.showOnPriceTable,
					massPerLiter		= fillType.massPerLiter,
					massFactor			= 1}

				local unitTypes = ucMod:getUnitTypesByGroup(ucFillType.group)
				for _, unitSet in ipairs(ucMod.unitSetsByIndex) do
					if unitSet.isFillTypeSet then
						ucFillType.units[unitSet.index] = getDefaultUnitByTypes(unitTypes)
					end
				end

				ucMod.fillTypesByIndex[ucFillType.index] = ucFillType
				ucMod.fillTypesByName[ucFillType.name] = ucFillType
			end
			fillType.showOnPriceTable = ucFillType.showOnPriceTable
		end
	end

	-- Make UnitConvert texts global
	for key, text in pairs(g_i18n.texts) do
		getfenv(0)["g_i18n"]:setText(key, text)
	end

	-- Setup configuration menu
	ucMod:setupConfigMenu()
end

function UnitConvert:saveConfig()
	local baseKey = "UnitConvert"
	local saveXML = createXMLFile(baseKey, ucMod.saveFile, baseKey)

	if xAssert(saveXML) then
		setXMLInt(saveXML, baseKey.."#version", ucMod.configVersion)
		setXMLString(saveXML, baseKey.."#lengthUnit", ucMod.unitsByIndex[ucMod.lengthUnit].name)
		setXMLString(saveXML, baseKey.."#weightUnit", ucMod.unitsByIndex[ucMod.weightUnit].name)
		setXMLString(saveXML, baseKey.."#areaUnit", ucMod.unitsByIndex[ucMod.areaUnit].name)
		setXMLString(saveXML, baseKey.."#yieldDisplay", ucMod.yieldDisplay)

		local idx = 0
		for _, unit in ipairs(ucMod.unitsByIndex) do
			local unitKey = string.format("%s.unit(%d)", baseKey, idx)
			setXMLString(saveXML, unitKey.."#name", unit.name)
			setXMLInt(saveXML, unitKey.."#precision", unit.precision)
			idx = idx + 1
		end

		idx = 0
		for _, fillTypeGroup in ipairs(ucMod.fillTypeGroupsByIndex) do
			if fillTypeGroup.showOnMenu then
				local fillTypeGroupKey = string.format("%s.fillTypeGroup(%d)", baseKey, idx)
				setXMLString(saveXML, fillTypeGroupKey.."#name", fillTypeGroup.name)
				setXMLInt(saveXML, fillTypeGroupKey.."#priority", fillTypeGroup.priority)
				for _, unitSet in ipairs(ucMod.unitSetsByIndex) do
					if fillTypeGroup.units[unitSet.index] then
						local unit = ucMod:getUnit(fillTypeGroup.units[unitSet.index])
						setXMLString(saveXML, fillTypeGroupKey.."#unit"..tostring(unitSet.index), unit.name)
					end
				end
				idx = idx + 1
			end
		end

		idx = 0
		for _, fillType in ipairs(ucMod.fillTypesByIndex) do
			if fillType.showOnMenu then
				local fillTypeKey = string.format("%s.fillType(%d)", baseKey, idx)
				setXMLString(saveXML, fillTypeKey.."#name", fillType.name)
				setXMLFloat(saveXML, fillTypeKey.."#massFactor", fillType.massFactor)
				for _, unitSet in ipairs(ucMod.unitSetsByIndex) do
					if fillType.units[unitSet.index] then
						local unit = ucMod:getUnit(fillType.units[unitSet.index])
						setXMLString(saveXML, fillTypeKey.."#unit"..tostring(unitSet.index), unit.name)
					end
				end
				idx = idx + 1
			end
		end
		saveXMLFile(saveXML)
		delete(saveXML)
	end
end

-- [MAIN PROGRAM] *************************************************************************************************************************************************************************************

function UnitConvert:loadMap(name)
	-- General constants
	ucMod.PRECISION_MAX = 4

	ucMod.MASS_FACTOR_MIN	= 0.50
	ucMod.MASS_FACTOR_MAX	= 5.00
	ucMod.MASS_FACTOR_STEP	= 0.25

	ucMod.YIELD_VEHICLE	= "vehicle"
	ucMod.YIELD_PRICE	= "price"
	ucMod.YIELD_OFF		= "off"

	-- Color constants
	ucMod.colors = {}
	ucMod.colors.white				= {1, 1, 1, 1}
	ucMod.colors.red				= {0.8069, 0.0097, 0.0097, 1}
	ucMod.colors.green				= {0.2122, 0.5271, 0.0307, 1}
	ucMod.colors.green2 			= {0.2705, 0.6514, 0.0802, 1}
	ucMod.colors.yellow				= {0.9301, 0.6404, 0.0439, 1}
	ucMod.colors.darkGrey			= {0.0284, 0.0284, 0.0284, 1}
	ucMod.colors.darkGrey2			= {0.0630, 0.0630, 0.0630, 1}
	ucMod.colors.darkGrey3			= {0.0194, 0.0194, 0.0194, 1}
	ucMod.colors.darkGrey7			= {0.0104, 0.0104, 0.0104, 1}
	ucMod.colors.black				= {0, 0, 0, 1}
	ucMod.colors.black2				= {0.0075, 0.0075, 0.0075, 1}
	ucMod.colors.textUnitColor		= {0.0865, 0.0865, 0.0865, 1}
	ucMod.colors.textUnitColor2		= {0.3365, 0.3365, 0.3365, 1}
	ucMod.colors.descriptionText4	= {0.6307, 0.6307, 0.6307, 1}

	-- Unit types
	ucMod.unitTypesByIndex = {}
	table.insert(ucMod.unitTypesByIndex, {name = "piece"})
	table.insert(ucMod.unitTypesByIndex, {name = "animal"})
	table.insert(ucMod.unitTypesByIndex, {name = "volume"})
	table.insert(ucMod.unitTypesByIndex, {name = "weight"})
	table.insert(ucMod.unitTypesByIndex, {name = "length"})
	table.insert(ucMod.unitTypesByIndex, {name = "area"})

	ucMod.unitTypesByName = {}
	for unitTypeIdx, unitType in pairs(ucMod.unitTypesByIndex) do
		unitType.index = unitTypeIdx
		unitType.nameI18N = getText("UC_unit_type_"..unitType.name)
		ucMod.unitTypesByName[unitType.name] = unitType
		ucMod["UNIT_TYPE_"..string.upper(unitType.name)] = unitType.index
	end
	ucMod.NUM_UNIT_TYPES = #ucMod.unitTypesByIndex

	-- Unit sets
	ucMod.unitSetsByIndex = {}
	table.insert(ucMod.unitSetsByIndex, {name = "vehicle", isFillTypeSet = true, isFillTypeGroupSet = true})
	table.insert(ucMod.unitSetsByIndex, {name = "storage", isFillTypeSet = true, isFillTypeGroupSet = true})
	table.insert(ucMod.unitSetsByIndex, {name = "animal", isFillTypeSet = true, isFillTypeGroupSet = true})
	table.insert(ucMod.unitSetsByIndex, {name = "price", isFillTypeSet = true})

	ucMod.unitSetsByName = {}
	for unitSetIdx, unitSet in pairs(ucMod.unitSetsByIndex) do
		unitSet.index = unitSetIdx
		unitSet.nameI18N = getText("UC_unit_group_"..unitSet.name)
		ucMod.unitSetsByName[unitSet.name] = unitSet
		ucMod["UNIT_SET_"..string.upper(unitSet.name)] = unitSet.index
	end
	ucMod.NUM_UNIT_SETS = #ucMod.unitSetsByIndex

	-- Configuration menu
	ucMod.configMenu = {}
	ucMod.configMenu.overlayId		= createImageOverlay(ucMod.modDir.."overlay/overlay.dds")
	ucMod.configMenu.arrowUpId		= createImageOverlay(ucMod.modDir.."overlay/arrowUp.dds")
	ucMod.configMenu.arrowDownId	= createImageOverlay(ucMod.modDir.."overlay/arrowDown.dds")
	ucMod.configMenu.textSize		= g_currentMission.helpBoxTextSize * 0.85
	ucMod.configMenu.marginX		= (0.0075 / g_screenAspectRatio)  * uiScale
	ucMod.configMenu.marginY		= 0.0075 * uiScale
	ucMod.configMenu.lineSpacingX	= (0.005 / g_screenAspectRatio) * uiScale
	ucMod.configMenu.lineSpacingY	= 0.0025 * uiScale

	ucMod.configMenu.colors = {}
	ucMod.configMenu.colors["titleBg"]			= ucMod.colors.black2
	ucMod.configMenu.colors["titleText"]		= ucMod.colors.white
	ucMod.configMenu.colors["itemBg"]			= ucMod.colors.darkGrey7
	ucMod.configMenu.colors["itemBgEven"]		= ucMod.colors.darkGrey3
	ucMod.configMenu.colors["itemBgOdd"]		= ucMod.colors.darkGrey
	ucMod.configMenu.colors["itemBgSelected"]	= ucMod.colors.darkGrey2
	ucMod.configMenu.colors["itemText"]			= ucMod.colors.descriptionText4
	ucMod.configMenu.colors["itemTextSelected"]	= ucMod.colors.green2

	ucMod.configMenu.pages = {}
	table.insert(ucMod.configMenu.pages, {name = "fillTypes"})
	table.insert(ucMod.configMenu.pages, {name = "fillTypeGroups"})
	table.insert(ucMod.configMenu.pages, {name = "other"})
	table.insert(ucMod.configMenu.pages, {name = "units"})

	for pageIdx, page in ipairs(ucMod.configMenu.pages) do
		page.index = pageIdx
		page.nameI18N = getText("UC_menu_page_"..page.name)
		ucMod.configMenu["PAGE_"..string.upper(page.name)] = page.index
	end

	ucMod.configMenu.NUM_PAGES = #ucMod.configMenu.pages
	ucMod.configMenu.NUM_VISIBLE_ROWS = math.floor(40 / uiScale)
	ucMod.configMenu.pageIndex = 1

	-- Seasons (or any mod that changes AnimalUtil data) compatibility
	ucMod.animalsByIndex = {}
	ucMod.animalsByName = {}
	for _, animal in pairs(AnimalUtil.animalIndexToDesc) do
		ucMod.animalsByIndex[animal.index] = {}

		ucMod.animalsByIndex[animal.index].name = animal.name
		ucMod.animalsByIndex[animal.index].index = animal.index

		ucMod.animalsByIndex[animal.index].hasWater = (animal.waterPerDay > 0)
		ucMod.animalsByIndex[animal.index].hasStraw = (animal.strawPerDay > 0)
		ucMod.animalsByIndex[animal.index].hasPallet = (animal.palletFillLevelPerDay > 0)
		ucMod.animalsByIndex[animal.index].hasManure = (animal.manurePerDay > 0)
		ucMod.animalsByIndex[animal.index].hasLiquidManure = (animal.liquidManurePerDay > 0)
		ucMod.animalsByIndex[animal.index].hasMilk = (animal.milkPerDay > 0)

		ucMod.animalsByName[animal.name] = ucMod.animalsByIndex[animal.index]
	end

	-- Bug fix #002
	ucMod:fixFillTypeMasses()
end

function UnitConvert:postLoadMapFinished(superFunc, ...)
	local ret = {superFunc(self, ...)}

--[[ Do most things here to prevent being overwritten by other mods --]]

	-- Load Configuration(s)
	ucMod:loadConfig()

	-- Setup help lines
	ucMod:setupHelpLines()

	-- InGameMenu - Prices Page
	g_inGameMenu.setupPriceTable	= Utils.overwrittenFunction(g_inGameMenu.setupPriceTable, ucMod.inGameMenuSetupPriceTable)
	g_inGameMenu.updatePriceTable	= Utils.overwrittenFunction(g_inGameMenu.updatePriceTable, ucMod.inGameMenuUpdatePriceTable)

	-- InGameMenu - Statistics Page
	g_inGameMenu.updateStatisticTable = Utils.overwrittenFunction(g_inGameMenu.updateStatisticTable, ucMod.inGameMenuUpdateStatisticTable)

	-- InGameMenu - Animals Page
	g_inGameMenu.getAnimalData = Utils.overwrittenFunction(g_inGameMenu.getAnimalData, ucMod.inGameMenuGetAnimalData)

	-- ShopScreen / ShopScreenConfig
	g_shopScreen.updateAttributes = Utils.overwrittenFunction(g_shopScreen.updateAttributes, ucMod.shopScreenUpdateAttributes)
	g_shopConfigScreen.updateData = Utils.overwrittenFunction(g_shopConfigScreen.updateData, ucMod.shopConfigScreenUpdateData)

	-- SiloDialog
	g_gui.showSiloDialog = Utils.overwrittenFunction(g_gui.showSiloDialog, ucMod.guiShowSiloDialog)

	-- FieldJobMissionScreen
	g_fieldJobMissionScreen.setData = Utils.overwrittenFunction(g_fieldJobMissionScreen.setData, ucMod.fieldJobMissionScreenSetData)

	-- BGA
	Bga.setFillLevel = Utils.overwrittenFunction(Bga.setFillLevel, ucMod.bgaSetFillLevel)

	-- BunkerSilo
	BunkerSilo.update = Utils.overwrittenFunction(BunkerSilo.update, ucMod.bunkerSiloUpdate)

	-- LiquidManureFillTrigger
	LiquidManureFillTrigger.update = Utils.overwrittenFunction(LiquidManureFillTrigger.update, ucMod.liquidManureFillTriggerUpdate)

	-- WeighStation
	WeighStation.updateWeight = Utils.overwrittenFunction(WeighStation.updateWeight, ucMod.weighStationUpdateWeight)

	-- VehicleHud
	g_currentMission.drawLevelBar = Utils.overwrittenFunction(g_currentMission.drawLevelBar, ucMod.currentMissionDrawLevelBar)
	VehicleHudUtils.setHudValue = Utils.overwrittenFunction(VehicleHudUtils.setHudValue, ucMod.vehicleHudUtilsSetHudValue)

	-- Additional Mods
	ucMod:loadThirdPartyAddOns()

	-- FOR TESTING PURPOSES
	if DebugTools then
		--DebugTools:searchTable(self, 4)
	end

	return unpack(ret)
end
BaseMission.loadMapFinished = Utils.overwrittenFunction(BaseMission.loadMapFinished, ucMod.postLoadMapFinished)

function UnitConvert:update(dt)
	if not g_gui:getIsGuiVisible() and not g_currentMission.controlledVehicle then
		if InputBinding.hasEvent(InputBinding.UC_OPEN_MENU) then
			ucMod.isMenuOpen = not ucMod.isMenuOpen
		end

		if ucMod.isMenuOpen then
			local cfgMenu = ucMod.configMenu

			g_currentMission:addHelpButtonText(getText("UC_menu_close"), InputBinding.UC_OPEN_MENU, nil, GS_PRIO_VERY_HIGH)

			g_currentMission:addHelpButtonText(getText("UC_menu_pageNext"), InputBinding.UC_MENU_PAGE_NEXT, nil, GS_PRIO_HIGH)
			g_currentMission:addHelpButtonText(getText("UC_menu_pagePrev"), InputBinding.UC_MENU_PAGE_PREV, nil, GS_PRIO_HIGH)
			if InputBinding.hasEvent(InputBinding.UC_MENU_PAGE_NEXT)
			or InputBinding.hasEvent(InputBinding.UC_MENU_PAGE_PREV)
			then
				local cfgMenuStep = 1
				if InputBinding.hasEvent(InputBinding.UC_MENU_PAGE_PREV) then cfgMenuStep = -1 end

				cfgMenu.pageIndex = cfgMenu.pageIndex + cfgMenuStep
				if cfgMenu.pageIndex < 1 then cfgMenu.pageIndex = cfgMenu.NUM_PAGES end
				if cfgMenu.pageIndex > cfgMenu.NUM_PAGES then cfgMenu.pageIndex = 1 end
			end
			local cfgMenuPage = cfgMenu.pages[cfgMenu.pageIndex]

			if InputBinding.hasEvent(InputBinding.UC_MENU_DOWN)
			or InputBinding.hasEvent(InputBinding.UC_MENU_UP)
			then
				if cfgMenuPage.startRow > 0 then
					local minRow = cfgMenuPage.startRow
					local maxRow = minRow + (cfgMenu.NUM_VISIBLE_ROWS - 1)

					local cfgMenuStep = 1
					if InputBinding.hasEvent(InputBinding.UC_MENU_UP) then cfgMenuStep = -1 end

					cfgMenuPage.listIndex = cfgMenuPage.listIndex + cfgMenuStep

					if cfgMenuPage.listIndex > cfgMenuPage.NUM_ROWS	then cfgMenuPage.listIndex = 1 end
					if cfgMenuPage.listIndex < 1 then cfgMenuPage.listIndex = cfgMenuPage.NUM_ROWS end

					if cfgMenuPage.listIndex > maxRow then cfgMenuPage.startRow = cfgMenuPage.listIndex - (cfgMenu.NUM_VISIBLE_ROWS - 1) end
					if cfgMenuPage.listIndex < cfgMenuPage.startRow then cfgMenuPage.startRow = cfgMenuPage.listIndex end
				end
			end

			if cfgMenu.pageIndex == cfgMenu.PAGE_FILLTYPEGROUPS then
				g_currentMission:addHelpButtonText(getText("UC_menu_itemUp"), InputBinding.UC_MENU_ITEM_UP, nil, GS_PRIO_HIGH)
				g_currentMission:addHelpButtonText(getText("UC_menu_itemDown"), InputBinding.UC_MENU_ITEM_DOWN, nil, GS_PRIO_HIGH)
				if InputBinding.hasEvent(InputBinding.UC_MENU_ITEM_UP)
				or InputBinding.hasEvent(InputBinding.UC_MENU_ITEM_DOWN)
				then
					if cfgMenuPage.startRow > 0 then
						local minRow = cfgMenuPage.startRow
						local maxRow = minRow + (cfgMenu.NUM_VISIBLE_ROWS - 1)

						local cfgMenuStep = 1
						if InputBinding.hasEvent(InputBinding.UC_MENU_ITEM_UP) then cfgMenuStep = -1 end

						local oldListIndex = cfgMenuPage.listIndex

						cfgMenuPage.listIndex = cfgMenuPage.listIndex + cfgMenuStep

						if cfgMenuPage.listIndex > cfgMenuPage.NUM_ROWS	then cfgMenuPage.listIndex = 1 end
						if cfgMenuPage.listIndex < 1 then cfgMenuPage.listIndex = cfgMenuPage.NUM_ROWS end

						if cfgMenuPage.listIndex > maxRow then cfgMenuPage.startRow = cfgMenuPage.listIndex - (cfgMenu.NUM_VISIBLE_ROWS - 1) end
						if cfgMenuPage.listIndex < cfgMenuPage.startRow then cfgMenuPage.startRow = cfgMenuPage.listIndex end

						local oldItem = table.remove(cfgMenuPage.listItems, oldListIndex)
						table.insert(cfgMenuPage.listItems, cfgMenuPage.listIndex, oldItem)

						for listIndex, listItem in ipairs(cfgMenuPage.listItems) do
							ucMod.fillTypeGroupsByPriority[listIndex] = ucMod.fillTypeGroupsByName[listItem.title]
							ucMod.fillTypeGroupsByPriority[listIndex].priority = listIndex
						end
						ucMod:saveConfig()
					end
				end
			end

			if InputBinding.hasEvent(InputBinding.UC_MENU_LEFT)
			or InputBinding.hasEvent(InputBinding.UC_MENU_RIGHT)
			then
				local cfgMenuStep = 1
				if InputBinding.hasEvent(InputBinding.UC_MENU_LEFT) then cfgMenuStep = -1 end

				cfgMenuPage.columnIndex = cfgMenuPage.columnIndex + cfgMenuStep
				if cfgMenuPage.columnIndex > cfgMenuPage.NUM_COLUMNS then cfgMenuPage.columnIndex = 1 end
				if cfgMenuPage.columnIndex < 1 then cfgMenuPage.columnIndex = cfgMenuPage.NUM_COLUMNS end
			end

			local currentRow = cfgMenuPage.listItems[cfgMenuPage.listIndex]
			local currentCol = currentRow.columns[cfgMenuPage.columnIndex]

			local menuSettingType = "Unit"
			if (cfgMenu.pageIndex == cfgMenu.PAGE_FILLTYPES and cfgMenuPage.columnIndex == cfgMenuPage.COLUMN_MASS_FACTOR) or cfgMenu.pageIndex == cfgMenu.PAGE_UNITS then
				menuSettingType = "Value"
			elseif cfgMenu.pageIndex == cfgMenu.PAGE_FILLTYPES and cfgMenuPage.columnIndex == cfgMenuPage.COLUMN_FILLTYPE_GROUP then
				menuSettingType = nil
			end

			if menuSettingType then
				g_currentMission:addHelpButtonText(getText("UC_menu_next"..menuSettingType), InputBinding.UC_MENU_VALUE_NEXT, nil, GS_PRIO_HIGH)
				g_currentMission:addHelpButtonText(getText("UC_menu_prev"..menuSettingType), InputBinding.UC_MENU_VALUE_PREV, nil, GS_PRIO_HIGH)
			end

			if InputBinding.hasEvent(InputBinding.UC_MENU_VALUE_NEXT)
			or InputBinding.hasEvent(InputBinding.UC_MENU_VALUE_PREV)
			then
				local cfgMenuStep = 1
				if InputBinding.hasEvent(InputBinding.UC_MENU_VALUE_PREV) then cfgMenuStep = -1 end

				currentCol.valueIndex = currentCol.valueIndex + cfgMenuStep
				if currentCol.valueIndex > currentCol.NUM_VALUES then currentCol.valueIndex = 1 end
				if currentCol.valueIndex < 1 then currentCol.valueIndex = currentCol.NUM_VALUES end

				local currentVal = currentCol.valuesByIndex[currentCol.valueIndex].name

				if cfgMenu.pageIndex == cfgMenu.PAGE_FILLTYPES or cfgMenuPage == cfgMenu.PAGE_FILLTYPEGROUPS then
					local target = 0
					if cfgMenu.pageIndex == cfgMenu.PAGE_FILLTYPES then
						target = ucMod:getFillType(currentRow.title)
					else
						target = ucMod:getFillTypeGroup(currentRow.title)
					end
					if cfgMenuPage.columnIndex >= cfgMenuPage.COLUMN_UNIT_SET_MIN and cfgMenuPage.columnIndex <= cfgMenuPage.COLUMN_UNIT_SET_MAX then
						local unitSetIdx = cfgMenuPage.columnIndex - (cfgMenuPage.COLUMN_UNIT_SET_MIN - 1)
						local unit = ucMod:getUnit(currentVal)
						if target.units[unitSetIdx] then
							target.units[unitSetIdx] = currentVal
						end
					else
						if cfgMenu.pageIndex == cfgMenu.PAGE_FILLTYPES then
							if cfgMenuPage.columnIndex == cfgMenuPage.COLUMN_MASS_FACTOR then
								local massFactor = ucMod:strToNum(currentVal) / 100
								target.massFactor = massFactor
							end
						end
					end
				elseif cfgMenu.pageIndex == cfgMenu.PAGE_OTHER then
					local unitType = ucMod:getUnitType(currentRow.title)
					local unit = ucMod:getUnit(currentVal)

					if unitType.index == ucMod.UNIT_TYPE_WEIGHT then
						ucMod.weightUnit = unit.index
					elseif unitType.index == ucMod.UNIT_TYPE_LENGTH then
						ucMod.lengthUnit = unit.index
					elseif unitType.index == ucMod.UNIT_TYPE_AREA then
						ucMod.areaUnit = unit.index
					end
				elseif cfgMenu.pageIndex == cfgMenu.PAGE_UNITS then
					local unit = ucMod:getUnit(currentRow.title)
					unit.precision = tonumber(currentVal)
				end
				ucMod:saveConfig()
			end

			if (cfgMenu.pageIndex == cfgMenu.PAGE_FILLTYPES or cfgMenu.pageIndex == cfgMenu.PAGE_FILLTYPEGROUPS) and (cfgMenuPage.columnIndex >= cfgMenuPage.COLUMN_UNIT_SET_MIN and cfgMenuPage.columnIndex <= cfgMenuPage.COLUMN_UNIT_SET_MAX) then
				g_currentMission:addHelpButtonText(getText"UC_menu_syncUnits", InputBinding.UC_MENU_SPECIAL, nil, GS_PRIO_HIGH)
				if InputBinding.hasEvent(InputBinding.UC_MENU_SPECIAL) then
					local currentVal = currentCol.valuesByIndex[currentCol.valueIndex].name
					local unit = ucMod:getUnit(currentVal)

					local target = 0
					if cfgMenu.pageIndex == cfgMenu.PAGE_FILLTYPES then
						target = ucMod:getFillType(currentRow.title)
					else
						target = ucMod:getFillTypeGroup(currentRow.title)
					end

					for _, unitSet in ipairs(ucMod.unitSetsByIndex) do
						local columnIdx = unitSet.index + (cfgMenuPage.COLUMN_UNIT_SET_MIN - 1)
						if columnIdx >= cfgMenuPage.COLUMN_UNIT_SET_MIN and columnIdx <= cfgMenuPage.COLUMN_UNIT_SET_MAX then
							if currentRow.columns[columnIdx] then
								local column = currentRow.columns[columnIdx]
								if column.valuesByName[currentVal] then
									column.valueIndex = column.valuesByName[currentVal].index
									if target.units[unitSet.index] then
										target.units[unitSet.index] = unit.index
									end
								end
							end
						end
						ucMod:saveConfig()
					end
				end
			elseif cfgMenu.pageIndex == cfgMenu.PAGE_FILLTYPES and cfgMenuPage.columnIndex == cfgMenuPage.COLUMN_MASS_FACTOR then
				g_currentMission:addHelpButtonText(getText("UC_menu_syncValues"), InputBinding.UC_MENU_SPECIAL, nil, GS_PRIO_HIGH)
				if InputBinding.hasEvent(InputBinding.UC_MENU_SPECIAL) then
					local currentVal = currentCol.valuesByIndex[currentCol.valueIndex].name
					local massFactor = ucMod:strToNum(currentVal) / 100

					for _, listItem in pairs(cfgMenuPage.listItems) do
						local fillType = ucMod:getFillType(listItem.title)
						local column = listItem.columns[cfgMenuPage.columnIndex]
						if column.valuesByName[currentVal] then
							column.valueIndex = column.valuesByName[currentVal].index
							fillType.massFactor = massFactor
						end
					end
					ucMod:saveConfig()
				end
			end
		else
			g_currentMission:addHelpButtonText(getText("UC_menu_open"), InputBinding.UC_OPEN_MENU, nil, GS_PRIO_NORMAL)
		end
	else
		ucMod.isMenuOpen = false
	end
end

function UnitConvert:draw()
	if ucMod.isMenuOpen then
		ucMod:drawConfigMenu()
	end
end

function UnitConvert:deleteMap() end
function UnitConvert:mouseEvent(posX, posY, isDown, isUp, button) end
function UnitConvert:keyEvent(unicode, sym, modifier, isDown) end

-- [INGAMEMENU OVERRIDES - PRICES PAGE] ***************************************************************************************************************************************************************

function UnitConvert:inGameMenuSetupPriceTable(superFunc)
	if not self.unitConvert then self.unitConvert = {} end

	if not self.unitConvert.priceTableInitialized then
		self.fillTypeUnits = {}
		self.fillTypeNames = {}
		for idx = 1, self.MAX_NUM_FILLTYPES do
			self.fillTypeUnits[idx] = self.fillLevels[idx]:clone(self.fillLevels[idx].parent)
			self.fillTypeUnits[idx]:setTextSize(getCorrectTextSize(self.fillLevels[idx].defaultTextSize * 0.65))
			self.fillTypeUnits[idx].textUpperCase = true
			self.fillTypeUnits[idx].textBold = true

			local fillTypeUnitPosY = self.fillTypeUnits[idx].parent.size[2] - self.fillTypeUnits[idx]:getTextHeight() - 0.002
			self.fillTypeUnits[idx]:setPosition(0, fillTypeUnitPosY)
			self.fillTypeUnits[idx]:setTextColor(unpack(ucMod.colors.red))

			self.fillTypeNames[idx] = self.fillLevels[idx]:clone(self.fillLevels[idx].parent)
			self.fillTypeNames[idx]:setTextSize(getCorrectTextSize(self.fillLevels[idx].defaultTextSize) * 0.65)
			self.fillTypeNames[idx].textUpperCase = true
			self.fillTypeNames[idx].textBold = true

			local fillTypeNamePosY = self.fillTypeNames[idx].parent.size[2] + 0.002
			self.fillTypeNames[idx]:setPosition(0, fillTypeNamePosY)
			self.fillTypeNames[idx]:setTextColor(unpack(ucMod.colors.green2))

			self.fillLevels[idx]:setPosition(0, 0.006)

			self.fillIcons[idx].position[2] = self.fillIcons[idx].position[2] + 0.005
			self.fillIcons[idx]:setPosition(unpack(self.fillIcons[idx].position))
		end

		for idx, page in pairs(self.pagingElement.pageMapping) do
			if page.id == InGameMenu.PAGE_PRICES then
				page.title = getText("UC_inGameMenu_prices")
				self.pageSelector.texts[idx] = page.title
				break
			end
		end

		local priceHeaderOffset = 0.01
		local priceHeaderBox = self.ingameMenuPriceHeaderBox
		local priceHeaderBoxFarm = priceHeaderBox.parent.elements[2]

		self.storageTitleElement = priceHeaderBoxFarm.elements[1].elements[1] -- TODO: find a easier way of finding these
		self.storageTitleElement:setText(getText("UC_inGameMenu_storageTotals"))

		for elementIdx, element in pairs(priceHeaderBox.elements) do
			element.size[2] = element.size[2] + priceHeaderOffset
			element:setSize(unpack(element.size))

			if elementIdx == 1 then
				element.elements[1]:updatePositionForOrigin("middleLeft")
				element.elements[1].position[2] = element.size[2] / 2
				element.elements[1]:setPosition(unpack(element.elements[1].position))
			end
		end

		priceHeaderBox.size[2] = priceHeaderBox.size[2] + priceHeaderOffset
		priceHeaderBox:setSize(unpack(priceHeaderBox.size))

		priceHeaderBoxFarm.position[2] = priceHeaderBoxFarm.position[2] - priceHeaderOffset
		priceHeaderBoxFarm:setPosition(unpack(priceHeaderBoxFarm.position))

		self.pricesList.position[2] = self.pricesList.position[2] - priceHeaderOffset
		self.pricesList:setPosition(unpack(self.pricesList.position))
		self.pricesList.visibleItems = self.pricesList.visibleItems - 1

		self.unitConvert.priceTableInitialized = true
	end

	local ret = {superFunc(self)}
	return unpack(ret)
end

function UnitConvert:inGameMenuUpdatePriceTable(superFunc, doNotUpdateSlider)
	local ret = {superFunc(self, doNotUpdateSlider)}

	local function getTextScale(text)
		local ret = 1
		if string.len(text) >= 1 then
			ret = ret - (string.len(text) / 100)
		end
		return ret
	end

	for idx = 1, InGameMenu.MAX_NUM_FILLTYPES do
		local priceFillType = self.priceFillTypes[self.startFillType + (idx - 1)]
		if priceFillType then
			local fillType = priceFillType.index
			local fillLevel = 0
			for _, storage in pairs(g_currentMission.storages) do
				fillLevel = fillLevel + Utils.getNoNil(storage.fillLevels[fillType], 0)
			end

			local fillUnit = ucMod:getFillUnit(fillType, ucMod.UNIT_SET_PRICE)
			fillLevel = ucMod:convertValue(fillLevel, fillUnit.index, fillType)
			fillLevel = ucMod:numToStr(fillLevel, fillUnit.precision, "ceil", true, false)

			local fillLevelTextSize = getCorrectTextSize(self.fillLevels[idx].defaultTextSize) * 0.95
			self.fillLevels[idx]:setTextSize(fillLevelTextSize * getTextScale(fillLevel))
			self.fillLevels[idx]:setText(fillLevel)

			self.fillTypeUnits[idx]:setText(fillUnit.nameI18N)
			self.fillTypeNames[idx]:setText(FillUtil.fillTypeIndexToDesc[fillType].nameI18N)

			for _, tipTriggerElement in ipairs(self.tipTriggers) do
				local priceElement = tipTriggerElement.prices[idx]
				local arrowElement = tipTriggerElement.arrows[idx]
				local parentElement = priceElement.parent

				arrowElement:setPosition(-0.0038, 0)

				for _, tipTrigger in ipairs(tipTriggerElement.tipTriggers) do
					if tipTrigger.acceptedFillTypes[fillType] then
						local fillPrice = tipTrigger:getEffectiveFillTypePrice(fillType)
						local fillPriceFactor = ucMod:convertValue(1, fillUnit.index, fillType)
						fillPrice = fillPrice / fillPriceFactor
						fillPrice = ucMod:numToStr(fillPrice, 2, "ceil", true, true)

						if string.len(fillPrice) > 6 then
						else
							priceElement:setTextSize(getCorrectTextSize(priceElement.defaultTextSize))
						end

						priceElement:setTextSize(getCorrectTextSize(priceElement.defaultTextSize) * getTextScale(fillPrice))
						priceElement:setText(string.format("%s %s", fillPrice, g_i18n:getCurrencySymbol(true)))
						break
					end
				end
				local priceElementX = 0 - arrowElement.size[1] - 0.0062
				local priceElementY = 0.0015 + ((parentElement.size[2] / 2) - (priceElement:getTextHeight() / 2))
				priceElement:setPosition(priceElementX, priceElementY)
			end
		end
	end

	return unpack(ret)
end

-- [INGAMEMENU OVERRIDES - STATISTICS PAGE] ***********************************************************************************************************************************************************

function UnitConvert:inGameMenuUpdateStatisticTable(superFunc)
	local ret = {superFunc(self)}

	for _, statistic in pairs(g_currentMission.missionStats.statisticData) do
		if statistic.valueSessionElem and statistic.valueTotalElem then
			statistic.valueTitleElem = statistic.valueTotalElem.parent.elements[1]

			local valueTitle = ""
			local valueSession = Utils.getNoNil(statistic.valueSession, "")
			local valueTotal = Utils.getNoNil(statistic.valueTotal, "")

			local fillType = FillUtil.FILLTYPE_UNKNOWN
			local statUnit = ucMod.UNIT_UNKNOWN
			local roundType = "floor"

			if string.find(statistic.name, getText("statistic_workedHectares")) then
				fillType = false
				statUnit = ucMod:getAreaUnit()
				valueTitle = string.format("%s [%s]", getText("UC_statistic_workedArea"), statUnit.nameI18NShort)
				if valueSession ~= "" then
					valueSession = g_currentMission.missionStats.statistics["workedHectares"].session
				end
				if valueTotal ~= "" then
					valueTotal = g_currentMission.missionStats.statistics["workedHectares"].total
				end
			elseif string.find(statistic.name, getText("statistic_cultivatedHectares")) then
				fillType = false
				statUnit = ucMod:getAreaUnit()
				valueTitle = string.format("%s [%s]", getText("UC_statistic_cultivatedArea"), statUnit.nameI18NShort)
				if valueSession ~= "" then
					valueSession = g_currentMission.missionStats.statistics["cultivatedHectares"].session
				end
				if valueTotal ~= "" then
					valueTotal = g_currentMission.missionStats.statistics["cultivatedHectares"].total
				end
			elseif string.find(statistic.name, getText("statistic_sownHectares")) then
				fillType = false
				statUnit = ucMod:getAreaUnit()
				valueTitle = string.format("%s [%s]", getText("UC_statistic_sownArea"), statUnit.nameI18NShort)
				if valueSession ~= "" then
					valueSession = g_currentMission.missionStats.statistics["sownHectares"].session
				end
				if valueTotal ~= "" then
					valueTotal = g_currentMission.missionStats.statistics["sownHectares"].total
				end
			elseif string.find(statistic.name, getText("statistic_fertilizedHectares")) then
				fillType = false
				statUnit = ucMod:getAreaUnit()
				valueTitle = string.format("%s [%s]", getText("UC_statistic_fertilizedArea"), statUnit.nameI18NShort)
				if valueSession ~= "" then
					valueSession = g_currentMission.missionStats.statistics["fertilizedHectares"].session
				end
				if valueTotal ~= "" then
					valueTotal = g_currentMission.missionStats.statistics["fertilizedHectares"].total
				end
			elseif string.find(statistic.name, getText("statistic_threshedHectares")) then
				fillType = false
				statUnit = ucMod:getAreaUnit()
				valueTitle = string.format("%s [%s]", getText("UC_statistic_threshedArea"), statUnit.nameI18NShort)
				if valueSession ~= "" then
					valueSession = g_currentMission.missionStats.statistics["threshedHectares"].session
				end
				if valueTotal ~= "" then
					valueTotal = g_currentMission.missionStats.statistics["threshedHectares"].total
				end
			elseif string.find(statistic.name, getText("statistic_fuelUsage")) then
				fillType = FillUtil.FILLTYPE_FUEL
				statUnit = ucMod:getFillUnit(fillType, ucMod.UNIT_SET_VEHICLE)
				valueTitle = string.format("%s [%s]", getText("statistic_fuelUsage"), statUnit.nameI18NShort)
				if valueSession ~= "" then
					valueSession = g_currentMission.missionStats.statistics["fuelUsage"].session
				end
				if valueTotal ~= "" then
					valueTotal = g_currentMission.missionStats.statistics["fuelUsage"].total
				end
			elseif string.find(statistic.name, getText("statistic_seedUsage")) then
				fillType = FillUtil.FILLTYPE_SEEDS
				statUnit = ucMod:getFillUnit(fillType, ucMod.UNIT_SET_VEHICLE)
				valueTitle = string.format("%s [%s]", getText("statistic_seedUsage"), statUnit.nameI18NShort)
				if valueSession ~= "" then
					valueSession = g_currentMission.missionStats.statistics["seedUsage"].session
				end
				if valueTotal ~= "" then
					valueTotal = g_currentMission.missionStats.statistics["seedUsage"].total
				end
			elseif string.find(statistic.name, getText("statistic_sprayUsage")) then
				-- Not quite sure what to do with this one, as there are multiple types of fertilizer (both liquid and solid)
				-- For now, using "fertilizer" as the template, based on the idea that you mix solid fertilizer with water to make liquid...
				fillType = FillUtil.FILLTYPE_FERTILIZER
				statUnit = ucMod:getFillUnit(fillType, ucMod.UNIT_SET_VEHICLE)
				valueTitle = string.format("%s [%s]", getText("statistic_sprayUsage"), statUnit.nameI18NShort)
				if valueSession ~= "" then
					valueSession = g_currentMission.missionStats.statistics["sprayUsage"].session
				end
				if valueTotal ~= "" then
					valueTotal = g_currentMission.missionStats.statistics["sprayUsage"].total
				end
			elseif string.find(statistic.name, getText("statistic_storageCapacity")) then
				statistic.valueTitleElem:setText(getText("UC_statistic_siloCapacity")) -- More accurate description
				if g_currentMission.farmSiloTrigger then
					local siloFillTypes = g_currentMission.farmSiloTrigger.fillTypes
					if getIsTable(siloFillTypes) and #siloFillTypes > 0 then
						fillType = ucMod:getHeaviestFillType(siloFillTypes)
						if fillType then
							statUnit = ucMod:getFillUnit(siloFillTypes, ucMod.UNIT_SET_STORAGE)
							roundType = "ceil"

							valueTitle = string.format("%s [%s]", getText("UC_statistic_siloCapacity"), statUnit.nameI18NShort)
							_, valueSession = g_currentMission.farmSiloTrigger:getAllFillLevels()
						end
					end
				end
			end

			if statUnit ~= ucMod.UNIT_UNKNOWN then
				statistic.valueTitleElem:setText(valueTitle)

				if getIsNumber(valueSession) then
					valueSession = ucMod:convertValue(valueSession, statUnit.index, fillType)
					valueSession = ucMod:numToStr(valueSession, statUnit.precision, roundType, true, false)
					statistic.valueSessionElem:setText(valueSession)
				end

				if getIsNumber(valueTotal) then
					valueTotal = ucMod:convertValue(valueTotal, statUnit.index, fillType)
					valueTotal = ucMod:numToStr(valueTotal, statUnit.precision, roundType, true, false)
					statistic.valueTotalElem:setText(valueTotal)
				end
			end
		end
	end
	return unpack(ret)
end

-- [INGAMEMENU FUNCTIONS - ANIMALS PAGE] **************************************************************************************************************************************************************

function UnitConvert:setAnimalAttribute(tbl, attrName, fillTypes, fillLevels, capacity, isPercent)
	if not getIsTable(tbl) or not getIsString(attrName) or not fillTypes or not fillLevels
	or (capacity and not getIsNumber(capacity))
	or (isPercent and not getIsBoolean(isPercent))
	then
		raiseError()
		return
	end

	if type(fillTypes) ~= "table" then fillTypes = ucMod:splitString(" ", fillTypes) end
	if type(fillLevels) ~= "table" then	fillLevels = ucMod:splitString(" ", fillLevels) end

	local unitSet = ucMod.UNIT_SET_ANIMAL
	local fillUnit = ucMod:getFillUnit(fillTypes, unitSet)

	local fillLevel = 0
	local ucFillLevel = 0
	capacity = Utils.getNoNil(capacity, 0)

	for idx = 1, #fillTypes do
		if xAssert(ucMod:getIsFillTypeValid(fillTypes[idx])) then
			fillLevels[idx] = Utils.getNoNil(fillLevels[idx], 0)

			fillLevel = fillLevel + fillLevels[idx]
			ucFillLevel = ucFillLevel + ucMod:convertValue(fillLevels[idx], fillUnit.index, fillTypes[idx])
		end
	end
	ucFillLevel = ucMod:numToStr(ucFillLevel, fillUnit.precision, "floor", true, false)

	local fillPercent = 0
	if isPercent and capacity > 0 then
		fillPercent = Utils.clamp(fillLevel / capacity, 0, 1)
	else
		isPercent = false
	end

	attrName = string.format("%s [%s]", attrName, fillUnit.nameI18NShort)
	if isPercent then
		ucFillLevel = string.format("%s (%s%%)", ucFillLevel, math.floor(fillPercent * 100))
		table.insert(tbl, {
			name = attrName,
			value = ucFillLevel,
			percent = fillPercent,
			isPercentValue = true})
	else
		table.insert(tbl, {
			name = attrName,
			value = ucFillLevel})
	end
end

-- [INGAMEMENU OVERRIDES - ANIMALS PAGE] **************************************************************************************************************************************************************

function UnitConvert:inGameMenuGetAnimalData(superFunc)
	local ret = {superFunc(self)}

	local animalAttributes = ret[1] -- First return value is the animals attributes table
	if animalAttributes then
		for _, attribute in pairs(animalAttributes) do
			local animalName = attribute.name
			if g_currentMission.husbandries then
				local animalData = g_currentMission.husbandries[animalName]
				if animalData and animalData.hasStatistics then
					attribute.attrSpecific = {}

					if ucMod.animalsByName[animalName].hasPallet then
						local fillType = animalData.palletFillType
						if ucMod:getIsFillTypeValid(fillType) then
							local capacity = math.huge -- Arbitrary number used to show fill level bar even when empty
							local fillLevel = 0

							if animalData.currentPallet then
								capacity = animalData.currentPallet.capacity
								fillLevel = animalData.currentPallet.fillLevel
							end

							local attrName = FillUtil.fillTypeIndexToDesc[fillType].nameI18N
							ucMod:setAnimalAttribute(attribute.attrSpecific, attrName, fillType, fillLevel, capacity, true)
						end
					end

					if ucMod.animalsByName[animalName].hasManure then
						local fillType = FillUtil.FILLTYPE_MANURE
						local fillLevel = Utils.getNoNil(animalData.manureFillLevel, 0)
						local attrName = FillUtil.fillTypeIndexToDesc[fillType].nameI18N
						ucMod:setAnimalAttribute(attribute.attrSpecific, attrName, fillType, fillLevel)
					end

					if ucMod.animalsByName[animalName].hasLiquidManure and animalData.liquidManureTrigger then
						local fillType = animalData.liquidManureTrigger.fillType
						if ucMod:getIsFillTypeValid(fillType) then
							local fillLevel = Utils.getNoNil(animalData.liquidManureTrigger.fillLevel, 0)
							local attrName = FillUtil.fillTypeIndexToDesc[fillType].nameI18N
							ucMod:setAnimalAttribute(attribute.attrSpecific, attrName, fillType, fillLevel)
						end
					end

					if ucMod.animalsByName[animalName].hasMilk then
						local fillType = FillUtil.FILLTYPE_MILK
						local fillLevel = Utils.getNoNil(animalData.fillLevelMilk, 0)
						local attrName = FillUtil.fillTypeIndexToDesc[fillType].nameI18N
						ucMod:setAnimalAttribute(attribute.attrSpecific, attrName, fillType, fillLevel)
					end

					if attribute.attrInput then
						local attrInput = {}
						for _, input in ipairs(attribute.attrInput) do
							if string.find(input.name, getText("statistic_cleanliness")) then
								table.insert(attrInput, input)
							end
						end

						if ucMod.animalsByName[animalName].hasWater then
							local fillType = FillUtil.FILLTYPE_WATER
							local capacity = animalData:getCapacity(fillType)
							if capacity and capacity > 0 then
								local fillLevel = Utils.getNoNil(animalData:getFillLevel(fillType), 0)
								local attrName = FillUtil.fillTypeIndexToDesc[fillType].nameI18N
								ucMod:setAnimalAttribute(attrInput, attrName, fillType, fillLevel, capacity, true)
							end
						end

						if ucMod.animalsByName[animalName].hasStraw then
							local fillType = FillUtil.FILLTYPE_STRAW
							local capacity = animalData:getCapacity(fillType)
							if capacity and capacity > 0 then
								local fillLevel = Utils.getNoNil(animalData:getFillLevel(fillType), 0)
								local attrName = FillUtil.fillTypeIndexToDesc[fillType].nameI18N
								ucMod:setAnimalAttribute(attrInput, attrName, fillType, fillLevel, capacity, true)
							end
						end

						local animalDesc = AnimalUtil.animals[animalName]
						if animalDesc and FillUtil.foodGroups[animalDesc.index] then
							for _, foodGroup in ipairs(FillUtil.foodGroups[animalDesc.index]) do
								local fillTypes = {}
								local fillLevels = {}
								local capacity = 0

								if foodGroup.fillTypes then
									for fillType, isValid in pairs(foodGroup.fillTypes) do
										if isValid and ucMod:getIsFillTypeValid(fillType) then
											table.insert(fillTypes, fillType)
											table.insert(fillLevels, Utils.getNoNil(animalData:getFillLevel(fillType), 0)) -- TODO: Make sure this is correct
											capacity = math.max(capacity, Utils.getNoNil(animalData:getCapacity(fillType, foodGroup), 0))
										end
									end
								end

								if #fillTypes > 0 then
									local attrName = foodGroup.nameI18N
									ucMod:setAnimalAttribute(attrInput, attrName, fillTypes, fillLevels, capacity, true)
								end
							end
						end
						attribute.attrInput = attrInput
					end
				end
			end
		end
	end
	return unpack(ret)
end

-- [SHOP SCREEN FUNCTIONS] ****************************************************************************************************************************************************************************

function UnitConvert:convertShopScreenAttributes(shopScreen, storeItem)
	if not getIsTable(shopScreen) or not getIsTable(storeItem) then
		raiseError()
		return
	end

	if not shopScreen.attrIcon or not storeItem.specs then
		return
	end

	for idx, attr in pairs(shopScreen.attrIcon) do
		if shopScreen.attrValue
		and shopScreen.attrValue[idx]
		and shopScreen.attrValue[idx].text
		and shopScreen.attrValue[idx].text ~= ""
		then
			-- Fuel Capacity
			if attr.profile == StoreItemsUtil.storeItemSpecsNameToDesc["fuel"].profile then
				local fillType = FillUtil.FILLTYPE_FUEL
				local fillUnit = ucMod:getFillUnit(fillType, ucMod.UNIT_SET_VEHICLE)
				local capacity = Utils.getNoNil(storeItem.specs.fuel, 0)

				capacity = ucMod:convertValue(capacity, fillUnit.index, fillType)
				capacity = ucMod:numToStr(capacity, fillUnit.precision, "ceil", true, false)

				shopScreen.attrValue[idx]:setText(string.format("%s %s", capacity, fillUnit.nameI18NShort))
			-- Fill Capacity
			elseif attr.profile == StoreItemsUtil.storeItemSpecsNameToDesc["capacity"].profile then
				if storeItem.specs.fillTypes and storeItem.specs.capacity then
					local fillTypes = FillUtil.FILLTYPE_UNKNOWN
					local capacity = Utils.getNoNil(storeItem.specs.capacity.value, 0)

					local categories = storeItem.specs.fillTypes.categoryNames
					if categories then
						fillTypes = ucMod:getFillTypesByCategories(categories)
					end

					if not fillTypes or fillTypes == FillUtil.FILLTYPE_UNKNOWN then
						-- Don"t convert anything with "bale" or "pieces" assigned as a unit
						if storeItem.specs.capacity.unit ~= "$l10n_unit_pieces" and storeItem.specs.capacity.unit ~= "$l10n_unit_bale" then
							fillTypes = storeItem.specs.fillTypes.fillTypes
						end
					end

					if fillTypes and fillTypes ~= FillUtil.FILLTYPE_UNKNOWN then
						local fillType = ucMod:getHeaviestFillType(fillTypes)
						if fillType then
							local fillUnit = ucMod:getFillUnit(fillTypes, ucMod.UNIT_SET_VEHICLE)

							capacity = ucMod:convertValue(capacity, fillUnit.index, fillType)
							capacity = ucMod:numToStr(capacity, fillUnit.precision, "ceil", true, false)

							shopScreen.attrValue[idx]:setText(string.format("%s %s", capacity, fillUnit.nameI18NShort))
						end
					end
				end
			-- Working Width
			elseif attr.profile == StoreItemsUtil.storeItemSpecsNameToDesc["workingWidth"].profile then
				local lengthUnit = ucMod:getLengthUnit()

				local workingWidth = Utils.getNoNil(storeItem.specs.workingWidth, 0)
				workingWidth = ucMod:convertValue(workingWidth, lengthUnit.index, false)
				workingWidth = ucMod:numToStr(workingWidth, lengthUnit.precision, "ceil", true, false)

				shopScreen.attrValue[idx]:setText(string.format("%s %s", workingWidth, lengthUnit.nameI18NShort))
			end
		end
	end
end

-- [SHOP SCREEN OVERRIDES] ****************************************************************************************************************************************************************************

function UnitConvert:shopScreenUpdateAttributes(superFunc)
	local ret = {superFunc(self)}

	if self.currentItemList and #self.currentItemList > 0 and self.currentItemIndex and self.currentItemIndex > 0 then
		ucMod:convertShopScreenAttributes(self, self.currentItemList[self.currentItemIndex])
	end

	return unpack(ret)
end

function UnitConvert:shopConfigScreenUpdateData(superFunc)
	local ret = {superFunc(self)}

	if self.storeItem then
		ucMod:convertShopScreenAttributes(self, self.storeItem)
	end

	return unpack(ret)
end

-- [SILO DIALOG FUNCTIONS] ****************************************************************************************************************************************************************************

function UnitConvert:setSiloDialogTitle(siloDialog)
	if siloDialog.stationName then
		siloDialog.dialogTitleElement:setText(string.upper(siloDialog.stationName))
	end

	local fillType = siloDialog.selectedFillType
	if ucMod:getIsFillTypeValid(fillType) then
		local fillUnit = ucMod:getFillUnit(fillType, ucMod.UNIT_SET_STORAGE)

		local capacity = Utils.getNoNil(siloDialog.capacity, 0)
		if capacity == math.huge then
			siloDialog.dialogTitleElement2:setText(string.upper(getText("UC_siloDialog_inifiniteCapacity")))
		else
			capacity = ucMod:convertValue(capacity, fillUnit.index, fillType)
			capacity = ucMod:numToStr(capacity, fillUnit.precision, "ceil", true, false)
			siloDialog.dialogTitleElement2:setText(string.upper(getText("UC_siloDialog_capacity", capacity, fillUnit.nameI18NShort)))
		end
	end
end

function UnitConvert:setSiloDialogText(siloDialog)
	siloDialog.siloText.position[1] = 0
	siloDialog.siloText:setPosition(unpack(siloDialog.siloText.position))
	siloDialog.siloText:setText(siloDialog.fillTypesElement.texts[siloDialog.selectedFillType])
	siloDialog.siloIcon:setVisible(false)
end

-- [SILO DIALOG OVERRIDES] ****************************************************************************************************************************************************************************

function UnitConvert:siloDialogOnClickFillTypes(superFunc, state)
	local ret = {superFunc(self, state)}

	ucMod:setSiloDialogTitle(self)
	ucMod:setSiloDialogText(self)

	return unpack(ret)
end

function UnitConvert:guiShowSiloDialog(superFunc, args)
	local ret = {superFunc(self, args)}

	if getIsTable(args) and getIsTable(args.target) then
		local siloDialog	= self.guis.SiloDialog.target
		local siloTrigger	= args.target

		-- Initialization
		if not siloDialog.unitConvert then siloDialog.unitConvert = {} end
		if not siloDialog.unitConvert.initialized then
			-- Allow special code to run when dialog fill-type is changed
			siloDialog.fillTypesElement.onClickCallback = Utils.overwrittenFunction(siloDialog.fillTypesElement.onClickCallback, ucMod.siloDialogOnClickFillTypes)

			-- Customize fillType text element
			siloDialog.fillTypesElement.textElement:setTextSize(getCorrectTextSize(siloDialog.fillTypesElement.textElement.defaultTextSize) * 0.9)
			--siloDialog.fillTypesElement.textElement.textBold = true

			-- Add an extra title element
			siloDialog.dialogTitleElement2 = siloDialog.dialogTitleElement:clone(siloDialog.dialogTitleElement.parent)

			-- Format title elements
			local dlgTitleText = siloDialog.dialogTitleElement
			local dlgTitleTextPosX, dlgTitleTextPosY = unpack(dlgTitleText.position)
			local dlgTitleTextHeight = dlgTitleText:getTextHeight()

			local dlgTitleText2 = siloDialog.dialogTitleElement2

			dlgTitleText2:setTextSize(dlgTitleText.textSize * 0.65)
			dlgTitleText2:setTextColor(unpack(ucMod.colors.green))

			local dlgTitleText2PosX, dlgTitleText2PosY = unpack(dlgTitleText2.position)
			local dlgTitleText2Height = dlgTitleText2:getTextHeight()

			local dlgTitleBg = dlgTitleText.parent
			local dlgTitleBgWidth, dlgTitleBgHeight = unpack(dlgTitleBg.size)

			dlgTitleTextPosY = dlgTitleBgHeight - dlgTitleTextHeight - 0.006
			dlgTitleText:setPosition(dlgTitleTextPosX, dlgTitleTextPosY)

			dlgTitleText2PosY = 0.001
			dlgTitleText2:setPosition(dlgTitleText2PosX, dlgTitleText2PosY)

			-- Initialization complete
			siloDialog.unitConvert.initialized = true
		end

		-- Add this for future reference
		siloDialog.stationName = siloTrigger.stationName

		if siloDialog.fillTypeMapping then
			for _, fillType in pairs(siloDialog.fillTypeMapping) do
				if ucMod:getIsFillTypeValid(fillType) then
					local fillUnit = ucMod:getFillUnit(fillType, ucMod.UNIT_SET_STORAGE)

					local fillLevel = Utils.getNoNil(siloDialog.fillLevels[fillType], 0)
					fillLevel = ucMod:convertValue(fillLevel, fillUnit.index, fillType)
					fillLevel = ucMod:numToStr(fillLevel, fillUnit.precision, "ceil", true, false)

					local fillText = FillUtil.fillTypeIndexToDesc[fillType].nameI18N
					siloDialog.fillTypesElement.texts[fillType] = string.format("%s: %s %s", fillText, fillLevel, fillUnit.nameI18NShort)
				end
			end
		end

		ucMod:setSiloDialogTitle(siloDialog)
		ucMod:setSiloDialogText(siloDialog)
	end

	return unpack(ret)
end

-- [FIELDJOBMISSIONSCREEN OVERRIDES] ******************************************************************************************************************************************************************

function UnitConvert:fieldJobMissionScreenSetData(superFunc, npcIndex, fieldNumber, map, area, price, jobReward, jobTitle, jobDesc, vehicles, jobTime)
	local ret = {superFunc(self, npcIndex, fieldNumber, map, area, price, jobReward, jobTitle, jobDesc, vehicles, jobTime)}

	if self.fieldArea and area then
		local ucArea = area
		local areaUnit = ucMod:getAreaUnit()
		ucArea = ucMod:convertValue(ucArea, areaUnit.index, false)
		ucArea = ucMod:numToStr(ucArea, areaUnit.precision, "ceil", true, false)
		self.fieldArea:setText(string.format("%s %s", ucArea, areaUnit.nameI18NShort))
	end

	return unpack(ret)
end

-- [BGA FUNCTIONS] ************************************************************************************************************************************************************************************

function UnitConvert:getBgaFillTypes(bga)
	local fillTypes = {}

	if bga.tipTriggers then
		for _, tipTrigger in pairs(bga.tipTriggers) do
			for acceptedFillType, isValid in pairs(tipTrigger.acceptedFillTypes) do
				if isValid and ucMod:getIsFillTypeValid(acceptedFillType) then
					if acceptedFillType ~= FillUtil.FILLTYPE_LIQUIDMANURE then -- Liquid manure is handled differently by the BGA
						local added = false
						for _, fillType in pairs(fillTypes) do
							if fillType == acceptedFillType then
								added = true
								break
							end
						end
						if not added then
							table.insert(fillTypes, acceptedFillType)
						end
					end
				end
			end
		end
	end

	if #fillTypes > 0 then return fillTypes end
end

-- [BGA OVERRIDES] ************************************************************************************************************************************************************************************

function UnitConvert:bgaSetFillLevel(superFunc, fillLevel, fillType, fillable, trigger)
	local oldUpdateDisplayNumbers = self.updateDisplayNumbers
	self.updateDisplayNumbers = function(...) end

	local ret = {superFunc(self, fillLevel, fillType, fillable, trigger)}

	self.updateDisplayNumbers = oldUpdateDisplayNumbers

	local bunkerFillLevel = Utils.getNoNil(self.bunkerFillLevel, 0)

	if fillType ~= FillUtil.FILLTYPE_LIQUIDMANURE then
		local fillTypes = ucMod:getBgaFillTypes(self)
		if fillTypes then
			local ucFillType = ucMod:getHeaviestFillType(fillTypes)
			if ucFillType then
				local fillUnit = ucMod:getFillUnit(fillTypes, ucMod.UNIT_SET_STORAGE)
				bunkerFillLevel = ucMod:convertValue(bunkerFillLevel, fillUnit.index, ucFillType)
			end
		end
	end

	self:updateDisplayNumbers(bunkerFillLevel)

	return unpack(ret)
end

-- [BUNKER SILO OVERRIDES] ****************************************************************************************************************************************************************************

function UnitConvert:bunkerSiloUpdate(superFunc, dt)
	local oldAddExtraPrintText = g_currentMission.addExtraPrintText
	g_currentMission.addExtraPrintText = function(...) end

	local ret = {superFunc(self, dt)}

	g_currentMission.addExtraPrintText = oldAddExtraPrintText

   if self:getCanInteract(true) then
        local fillType = self.inputFillType
        if self.state == BunkerSilo.STATE_CLOSED or self.state == BunkerSilo.STATE_FERMENTED or self.state == BunkerSilo.STATE_DRAIN then
            fillType = self.outputFillType
        end

		if ucMod:getIsFillTypeValid(fillType) then
			local fillUnit = ucMod:getFillUnit(fillType, ucMod.UNIT_SET_STORAGE)

			local fillLevel = Utils.getNoNil(self.fillLevel, 0)
			fillLevel = ucMod:convertValue(fillLevel, fillUnit.index, fillType)
			fillLevel = ucMod:numToStr(fillLevel, fillUnit.precision, "ceil", true, false)

			local fillTypeText = FillUtil.fillTypeIndexToDesc[fillType].nameI18N
			local fillUnitText = fillUnit.nameI18NShort

			if self.state == BunkerSilo.STATE_FILL then
				g_currentMission:addExtraPrintText(string.format("%s: %s [%s]", fillTypeText, fillLevel, fillUnitText))
				g_currentMission:addExtraPrintText(getText("info_compacting")..string.format(" %d%%", self.compactedPercent))
			elseif self.state == BunkerSilo.STATE_CLOSED or self.state == BunkerSilo.STATE_FERMENTED then
				g_currentMission:addExtraPrintText(getText("info_fermenting")..string.format(" (%s): %d%%", fillTypeText, self.fermentingPercent))
			elseif self.state == BunkerSilo.STATE_DRAIN then
				g_currentMission:addExtraPrintText(string.format("%s: %s [%s]", fillTypeText, fillLevel, fillUnitText))
			end
		end
    end

	return unpack(ret)
end

-- [LIQUIDMANUREFILLTRIGGER OVERRIDES] ********************************************************************************************************************************************************************

function UnitConvert:liquidManureFillTriggerUpdate(superFunc, dt)
	local oldAddExtraPrintText = g_currentMission.addExtraPrintText
	g_currentMission.addExtraPrintText = function(...) end

	local ret = {superFunc(self, dt)}

	g_currentMission.addExtraPrintText = oldAddExtraPrintText

    if self:getShowInfo() then
		local fillType = self.fillType
		if ucMod:getIsFillTypeValid(fillType) then
			local fillUnit = ucMod:getFillUnit(fillType, ucMod.UNIT_SET_STORAGE)
			local fillLevel = Utils.getNoNil(self.fillLevel, 0)
			local capacity = Utils.getNoNil(self.capacity, 0)

			fillLevel = Utils.clamp(fillLevel, 0, capacity)

			local fillPercent = 0
			if capacity > 0 then
				fillPercent = fillLevel / capacity
			end

			fillLevel = ucMod:convertValue(fillLevel, fillUnit.index, fillType)
			fillLevel = ucMod:numToStr(fillLevel, fillUnit.precision, "ceil", true, false)

			local fillTypeText = FillUtil.fillTypeIndexToDesc[fillType].nameI18N
			g_currentMission:addExtraPrintText(string.format("%s [%s]: %s (%s%%)", fillTypeText, fillUnit.nameI18NShort, fillLevel, math.floor(fillPercent * 100)))
		end
    end

	return unpack(ret)
end

-- [WEIGHSTATION OVERRIDES] ***************************************************************************************************************************************************************************

function UnitConvert:weighStationUpdateWeight(superFunc)
	local totalMass = 0
	local fillMassOldTotal = 0
	local fillMassNewTotal = 0

	for vehicle, _ in pairs(self.triggerVehicles) do
		local fillMassOld, fillMassNew = ucMod:getVehicleFillMass(vehicle)

		totalMass = totalMass + vehicle:getTotalMass()
		fillMassOldTotal = fillMassOldTotal + fillMassOld
		fillMassNewTotal = fillMassNewTotal + fillMassNew
	end

	local vehicleMass = totalMass - fillMassOldTotal
	totalMass = vehicleMass + fillMassNewTotal

	local weightUnit = ucMod:getWeightUnit()
	totalMass = ucMod:convertValue(totalMass, weightUnit.index, false)

	self:updateDisplayNumbers(totalMass)
end

-- [VEHICLE HUD OVERRIDES] ****************************************************************************************************************************************************************************

function UnitConvert:currentMissionDrawLevelBar(superFunc, levelBar, value, text, textColor, textUnit, textUnitColor, icon)
	local vehicle = self.controlledVehicle

	if Utils.areListsEqual(textUnitColor, ucMod.colors.textUnitColor) then
		textUnitColor = ucMod.colors.textUnitColor2
	end

	local fillType = FillUtil.FILLTYPE_UNKNOWN
	if icon == self.fuelLevelIconOverlay then
		fillType = FillUtil.FILLTYPE_FUEL
	else
		if self.fillTypeOverlays then
			for idx, fillTypeOverlay in pairs(self.fillTypeOverlays) do
				if icon == fillTypeOverlay then
					fillType = idx
					break
				end
			end
		end
	end

	local textFillType = ""
	local textUnit2 = ""

	if ucMod:getIsFillTypeValid(fillType) then
		local fillLevel = ucMod:strToNum(text)
		if fillLevel then
			local fillUnit = ucMod:getFillUnit(fillType, ucMod.UNIT_SET_VEHICLE)
			fillLevel = ucMod:convertValue(fillLevel, fillUnit.index, fillType)
			fillLevel = ucMod:numToStr(fillLevel, fillUnit.precision, "ceil", true, false)

			text = fillLevel

			if icon == self.fuelLevelIconOverlay then
				textUnit = fillUnit.nameI18N
				if vehicle and vehicle.unitConvert then
					local fuelUsage = vehicle.unitConvert.fuelUsage
					if fuelUsage and fuelUsage >= 0 and vehicle.isMotorStarted then
						fuelUsage = ucMod:convertValue(fuelUsage, fillUnit.index, fillType)
						fuelUsage = ucMod:numToStr(fuelUsage, fillUnit.precision, "ceil", true, false)
						textUnit2 = getText("UC_vehicleHud_fuelUsage", fuelUsage, fillUnit.nameI18NShort)
					end
				end
			else
				textFillType = FillUtil.fillTypeIndexToDesc[fillType].nameI18N
				textUnit2 = fillUnit.nameI18N
				if vehicle and vehicle.unitConvert then
					local vehicleYield = vehicle.unitConvert.yield
					if not vehicleYield or vehicle.unitConvert.outputFillType ~= fillType then
						for _, implement in pairs(vehicle.attachedImplements) do
							implement = implement.object
							if implement.unitConvert and implement.unitConvert.yield and implement.unitConvert.outputFillType == fillType then
								vehicleYield = implement.unitConvert.yield
								break
							end
						end
					end
					if vehicleYield and vehicleYield > 0 then
						if ucMod.yieldDisplay ~= ucMod.YIELD_OFF then
							local unitSet = ucMod.UNIT_SET_VEHICLE
							if ucMod.yieldDisplay == ucMod.YIELD_PRICE then
								unitSet = ucMod.UNIT_SET_PRICE
							end

							local areaUnit = ucMod:getAreaUnit()
							local yieldUnit = ucMod:getFillUnit(fillType, unitSet)

							vehicleYield = ucMod:convertValue(vehicleYield, yieldUnit.index, fillType)
							vehicleYield = vehicleYield / areaUnit.conversionFactor

							vehicleYield = ucMod:numToStr(vehicleYield, yieldUnit.precision, "ceil", true, false)
							textUnit2 = getText("UC_vehicleHud_yield", vehicleYield, yieldUnit.nameI18NShort, areaUnit.nameI18NShort)
						end
					end
				end
			end
		end
	end

	local oldIconRender = function(...) end
	if icon and levelBar then
		oldIconRender = icon.render
		icon.render = function(...) end
	end

	local ret = {superFunc(self, levelBar, value, text, textColor, textUnit, textUnitColor, icon)}

	if icon and levelBar then
		icon.render = oldIconRender
		icon:setPosition(icon.x, (levelBar.y - 0.0005))
		icon:render()
	end

	if ucMod:getIsFillTypeValid(fillType) then
		setTextColor(unpack(textUnitColor))
		setTextAlignment(RenderText.ALIGN_RIGHT)

		local textUnit2Size = self.levelTextSmallTextSize * 0.9
		local textUnit2PosX = levelBar.x + levelBar.width
		local textUnit2PosY = levelBar.y - getTextHeight(textUnit2Size, textUnit2) - 0.0005

		renderText(textUnit2PosX, textUnit2PosY, textUnit2Size, textUnit2)

		setTextColor(unpack(ucMod.colors.white))
		setTextAlignment(RenderText.ALIGN_LEFT)

		local textFillTypeSize = textUnit2Size
		local textFillTypePosX = icon.x
		local textFillTypePosY = textUnit2PosY

		renderText(textFillTypePosX, textFillTypePosY, textFillTypeSize, textFillType)

		setTextColor(1, 1, 1, 1)
	end

	return unpack(ret)
end

function UnitConvert:vehicleHudUtilsSetHudValue(superFunc, huds, value, maxValue)
	local doUpdate = false
	for _, hud in pairs(huds) do
		if hud.numbers then
			doUpdate = true
			break
		end
	end

	if doUpdate then
		local fillType = FillUtil.FILLTYPE_UNKNOWN
		if (self.fuelFillLevelHud and huds == self.fuelFillLevelHud) or (self.fuelUsageHud and huds == self.fuelUsageHud) then
			fillType = FillUtil.FILLTYPE_FUEL
		elseif self.workedHectaresHud and huds == self.workedHectaresHud then
			local areaUnit = ucMod:getAreaUnit()
			workedHectares = Utils.getNoNil(value, 0)
			value = ucMod:convertValue(workedHectares, areaUnit.index, false)
		else
			if self.fillUnits then
				for _, fillUnit in pairs(self.fillUnits) do
					if fillUnit.fillLevelHud and huds == fillUnit.fillLevelHud then
						fillType = fillUnit.lastValidFillType
						break
					end
				end
			end
		end

		if ucMod:getIsFillTypeValid(fillType) then
			local fillLevel = Utils.getNoNil(value, 0)
			local capacity = Utils.getNoNil(maxValue, fillLevel)

			local fillUnit = ucMod:getFillUnit(fillType, ucMod.UNIT_SET_VEHICLE)

			value = ucMod:convertValue(fillLevel, fillUnit.index, fillType)
			maxValue = ucMod:convertValue(capacity, fillUnit.index, fillType)

			value = Utils.clamp(value, 0, maxValue)
		end
	end

	local ret = {superFunc(self, huds, value, maxValue)}
	return unpack(ret)
end

-- [VEHICLE OVERRIDES] ********************************************************************************************************************************************************************************

function UnitConvert:vehicleAddCutterArea(superFunc, cutter, area, realArea, inputFruitType, outputFillType)
	local ret = {superFunc(self, cutter, area, realArea, inputFruitType, outputFillType)}

	local yield = 0

	if self.unitConvert.updateYield then
		if ucMod:getIsFillTypeValid(outputFillType) then
			local literPerSqm = 1
			if ucMod:getIsFruitTypeValid(inputFruitType) then
				literPerSqm = FruitUtil.fruitIndexToDesc[inputFruitType].literPerSqm
			end

			local yieldMultiplier = 0
			if realArea and realArea > 0 then
				yieldMultiplier = (area / realArea) * self.threshingScale
			end

			yield = literPerSqm * yieldMultiplier
			yield = yield * 10000 --Convert to hectares
		end
	end

	self.unitConvert.yield = yield
	self.unitConvert.inputFruitType = inputFruitType
	self.unitConvert.outputFillType = outputFillType

	return unpack(ret)
end

function UnitConvert:vehicleUpdate(superFunc, dt)
	local ret = {superFunc(self, dt)}

	-- Initialization
	if self.unitConvert and not self.unitConvert.isInitialized then
		if self.fuelFillLevel and Utils.getNoNil(self.fuelCapacity, 0) > 0 then
			self.unitConvert.fuelUsage = 0
		end

		if SpecializationUtil.hasSpecialization(Combine, self.specializations) and self.addCutterArea then
			self.addCutterArea = Utils.overwrittenFunction(self.addCutterArea, ucMod.vehicleAddCutterArea)
			self.unitConvert.yield = 0
		end

		self.unitConvert.isInitialized = true
	end

	if not self.attacherVehicle then
		if self:getIsActive() and self:getIsActiveForInput(false, false) then
			local showYieldToggle = false
			if self.unitConvert and self.unitConvert.yield then
				showYieldToggle = true
			else
				for _, implement in pairs(self.attachedImplements) do
					 implement = implement.object
					 if implement and implement.unitConvert and implement.unitConvert.yield then
						showYieldToggle = true
						break
					end
				end
			end
			if showYieldToggle then
				g_currentMission:addHelpButtonText(getText("UC_help_toggleYield"), InputBinding.UC_TOGGLE_YIELD, nil, GS_PRIO_VERY_LOW)
				g_currentMission:addExtraPrintText(getText("UC_help_yieldDisplay", ucMod.yieldDisplay))
				if InputBinding.hasEvent(InputBinding.UC_TOGGLE_YIELD) then
					if ucMod.yieldDisplay == ucMod.YIELD_OFF then ucMod.yieldDisplay = ucMod.YIELD_VEHICLE
					elseif ucMod.yieldDisplay == ucMod.YIELD_VEHICLE then ucMod.yieldDisplay = ucMod.YIELD_PRICE
					else ucMod.yieldDisplay = ucMod.YIELD_OFF
					end

					ucMod:saveConfig()
				end
			end
		end
	end

	return unpack(ret)
end
Vehicle.update = Utils.overwrittenFunction(Vehicle.update, ucMod.vehicleUpdate)

function UnitConvert:vehicleUpdateTick(superFunc, dt)
	local ret = {superFunc(self, dt)}

	if self.isServer then
		if self.unitConvert then
			self.unitConvert.updateTickDT = Utils.getNoNil(self.unitConvert.updateTickDT, 0) + dt
			if self.unitConvert.updateTickDT > 1000 then
				-- Fuel Usage
				if self.unitConvert.fuelUsage then
					self.unitConvert.lastFuelFillLevel = Utils.getNoNil(self.unitConvert.lastFuelFillLevel, self.fuelFillLevel)

					local fuelUsage = math.max(0, (self.unitConvert.lastFuelFillLevel - self.fuelFillLevel))
					self.unitConvert.fuelUsage = (fuelUsage / self.unitConvert.updateTickDT) * 1000 * 60 * 60
					self.unitConvert.lastFuelFillLevel = self.fuelFillLevel
				end

				-- Yield
				if self.unitConvert.yield and self.lastAreaZeroTime then
					self.unitConvert.updateYield = true
					if self.lastAreaZeroTime > 1000 then
						self.unitConvert.updateYield = false
						self.unitConvert.yield = 0
					end
				end

				self.unitConvert.updateTickDT = 0
			end
		end
	end

	return unpack(ret)
end
Vehicle.updateTick = Utils.overwrittenFunction(Vehicle.updateTick, ucMod.vehicleUpdateTick)

function UnitConvert:vehicleReadStream(superFunc, streamId, connection)
	local ret = {superFunc(self, streamId, connection)}
	return unpack(ret)
end
--Vehicle.readStream = Utils.overwrittenFunction(Vehicle.readStream, ucMod.vehicleReadStream)

function UnitConvert:vehicleWriteStream(superFunc, streamId, connection)
	local ret = {superFunc(self, streamId, connection)}
	return unpack(ret)
end
--Vehicle.writeStream = Utils.overwrittenFunction(Vehicle.writeStream, ucMod.vehicleWriteStream)

function UnitConvert:vehicleReadUpdateStream(superFunc, streamId, timeStamp, connection)
	local ret = {superFunc(self, streamId, timeStamp, connection)}

	if connection.isServer then
		local fuelUsagePct		= streamReadUIntN(streamId, 14)
		local yieldPct			= streamReadUInt16(streamId)
		local inputFruitType	= streamReadUInt8(streamId)
		local outputFillType	= streamReadUInt8(streamId)

		if self.unitConvert then
			if self.unitConvert.fuelUsage then
				fuelUsagePct = fuelUsagePct / 16383

				local fuelUsage = Utils.clamp((self.fuelCapacity * fuelUsagePct), 0, self.fuelCapacity)
				self.unitConvert.fuelUsage = fuelUsage
			end

			if self.unitConvert.yield then
				local yield = 0
				if yieldPct > 0 and ucMod:getIsFruitTypeValid(inputFruitType) then
					local fruitType = FruitUtil.fruitIndexToDesc[inputFruitType]
					yieldPct = yieldPct / 10000 -- Convert percentage to decimal (100.00% / 10000 = 1)
					yield = fruitType.literPerSqm * yieldPct * 10000  -- Convert to hectares (x 10000)
				end
				self.unitConvert.yield = yield
				self.unitConvert.inputFruitType = inputFruitType
				self.unitConvert.outputFillType = outputFillType
			end
		end
	end

	return unpack(ret)
end
Vehicle.readUpdateStream = Utils.overwrittenFunction(Vehicle.readUpdateStream, ucMod.vehicleReadUpdateStream)

function UnitConvert:vehicleWriteUpdateStream(superFunc, streamId, connection, dirtyMask)
	local ret = {superFunc(self, streamId, connection, dirtyMask)}

	if not connection.isServer then
		local fuelUsagePct = 0
		local yieldPct = 0
		local inputFruitType = FruitUtil.FRUITTYPE_UNKNOWN
		local outputFillType = FillUtil.FILLTYPE_UNKNOWN

		if self.unitConvert then
			if self.unitConvert.fuelUsage then
				fuelUsagePct = Utils.clamp((self.unitConvert.fuelUsage / self.fuelCapacity), 0, 1)
				fuelUsagePct = math.floor(fuelUsagePct * 16383)
			end

			if self.unitConvert.yield then
				inputFruitType = Utils.getNoNil(self.unitConvert.inputFruitType, inputFruitType)
				outputFillType = Utils.getNoNil(self.unitConvert.outputFillType, outputFillType)

				if self.unitConvert.yield > 0 and ucMod:getIsFruitTypeValid(inputFruitType) then
					local fruitType = FruitUtil.fruitIndexToDesc[inputFruitType]
					yieldPct = (self.unitConvert.yield / 10000) / fruitType.literPerSqm
					yieldPct = math.floor(Utils.clamp(yieldPct, 0, 6) * 10000) -- Up to 6x normal yield (6 = 600% * 10000 = 60000 < 65535 = 16 bit)
				end
			end
		end

		streamWriteUIntN(streamId, fuelUsagePct, 14)
		streamWriteUInt16(streamId, yieldPct)
		streamWriteUInt8(streamId, inputFruitType)
		streamWriteUInt8(streamId, outputFillType)
	end

	return unpack(ret)
end
Vehicle.writeUpdateStream = Utils.overwrittenFunction(Vehicle.writeUpdateStream, ucMod.vehicleWriteUpdateStream)

function UnitConvert:vehicleLoad(superFunc, ...)
	local ret = {superFunc(self, ...)}

	if not self.unitConvert then self.unitConvert = {} end

	return unpack(ret)
end
Vehicle.load = Utils.overwrittenFunction(Vehicle.load, ucMod.vehicleLoad)

-- [BUG FIXES] ****************************************************************************************************************************************************************************************

-- #001: Game still shows "start filling" text if trailer is only slightly filled and silo is empty.
	-- This may be something they implemented with the "fillTypeChangeThreshold" variable for fillable vehicles.
	-- If so, I couldn"t find a way to actually override a fillType at a silo in-game.

function UnitConvert:siloTriggerGetIsActivatable(superFunc)
	if self.isEnabled then
		if self.isFilling then
			if self.siloTrailer:getRootAttacherVehicle() == g_currentMission.controlledVehicle then
				return true
			end
		else
			if self.siloTrailer and self.activeTriggers >= 4 then
				local trailer = self.siloTrailer
				if trailer:getRootAttacherVehicle() == g_currentMission.controlledVehicle and trailer:getAllowFillFromAir() then
					if trailer:getFillLevel() == 0 then
						return true
					else
						local fillTypes = trailer:getCurrentFillTypes()
						for _,fillType in pairs(fillTypes) do
							if self.fillTypes[fillType] and trailer:getFillLevel(fillType) < trailer:getCapacity() and self:getFillLevel(fillType) > 0 then
								return true
							end
						end
					end
				end
			end
		end
	end
	return false
end
if isDev then
	SiloTrigger.getIsActivatable = Utils.overwrittenFunction(SiloTrigger.getIsActivatable, ucMod.siloTriggerGetIsActivatable)
end

-- #002: Not an actual bug, but some fillTypes have either 0 or a very high number (i.e. 1.000000) as a mass
	-- This may cause UnitConvert to display crazy numbers (or divide by zero errors) in certain cases

function UnitConvert:fixFillTypeMasses()
	-- Adjusting these fillTypes, simply because their mass is zero (which could cause errors)
	FillUtil.fillTypeNameToDesc["powerFood"].massPerLiter	= 0.000150 -- same as forage
	FillUtil.fillTypeNameToDesc["tarp"].massPerLiter		= 0.000005
end

addModEventListener(UnitConvert)
